import unittest
import mock

from lxml import etree
from ncclient.operations.rpc import RPCReply

from pyhpecw7.comware import HPCOM7, NCTimeoutError, ConnectionClosedError, NCError,\
    ConnectionAuthenticationError, ConnectionSSHError, ConnectionUnkownHostError,\
    ConnectionError, LockConflictError, UnlockConflictError, NcTransErrors, NcOpErrors, RPCError,\
    socket

class HPCOM7TestCase(unittest.TestCase):


    def setUp(self):
        self.device = HPCOM7(host='host', username='user', password='pass')
        self.device.connection = mock.MagicMock()
        self.device.connection.connected = True

    def test_init(self):
        self.assertEqual(self.device.host, 'host')
        self.assertEqual(self.device.username, 'user')
        self.assertEqual(self.device.password, 'pass')
        self.assertEqual(self.device.port, 830)
        self.assertEqual(self.device.timeout, 30)
        self.assertEqual(self.device.staged, [])
        self.assertEqual(self.device._locked, False)

    @mock.patch('pyhpecw7.comware.manager', autospec=True)
    def test_open(self, mock_manager):
        self.device.open()
        mock_manager.connect.assert_called_with(allow_agent=True,
                                                device_params={'name': 'hpcomware'},
                                                host='host',
                                                hostkey_verify=False,
                                                look_for_keys=False,
                                                password='pass',
                                                port=830,
                                                ssh_config=None,
                                                timeout=30,
                                                username='user')


    @mock.patch('pyhpecw7.comware.manager', autospec=True)
    def test_open_auth_error(self, mock_manager):
        mock_manager.connect.side_effect = NcTransErrors.AuthenticationError
        with self.assertRaises(ConnectionAuthenticationError):
            self.device.open()

    @mock.patch('pyhpecw7.comware.manager', autospec=True)
    def test_open_ssh_error(self, mock_manager):
        mock_manager.connect.side_effect = NcTransErrors.SSHError
        with self.assertRaises(ConnectionSSHError):
            self.device.open()

    @mock.patch('pyhpecw7.comware.manager', autospec=True)
    def test_open_dns_error(self, mock_manager):
        mock_manager.connect.side_effect = socket.gaierror
        with self.assertRaises(ConnectionUnkownHostError):
            self.device.open()

    @mock.patch('pyhpecw7.comware.manager', autospec=True)
    def test_open_ncclient_extensions_error(self, mock_manager):
        mock_manager.connect.side_effect = ImportError
        with self.assertRaisesRegexp(ImportError, 'comware extensions'):
            self.device.open()

    @mock.patch('pyhpecw7.comware.manager', autospec=True)
    def test_open_generic_error(self, mock_manager):
        mock_manager.connect.side_effect = Exception
        with self.assertRaises(ConnectionError):
            self.device.open()

    @mock.patch('pyhpecw7.comware.Facts')
    def test_facts(self, mock_facts):
        mock_facts.return_value.facts = {'a': 1}

        result = self.device.facts
        self.assertEqual(result, {'a': 1})

    @mock.patch('pyhpecw7.comware.Facts')
    def test_facts_no_connection(self, mock_facts):
        self.device.connection.connected = False
        mock_facts.return_value.facts = {'a': 1}

        result = self.device.facts
        self.assertEqual(result, None)

    def test_connected(self):
        self.assertEqual(self.device.connected, True)

    def test_connected_false(self):
        self.device.connection.connected = False
        self.assertEqual(self.device.connected, False)

    def test_connected_not_opened(self):
        del self.device.connection
        self.assertEqual(self.device.connected, False)

    def test_close(self):
        self.device.close()
        self.device.connection.close_session.assert_called_with()

    def test_close_timeout_error(self):
        self.device.connection.close_session.side_effect = NcOpErrors.TimeoutExpiredError

        with self.assertRaises(NCTimeoutError):
            self.device.close()

    def test_stage_config(self):
        for cfg_type in ['edit_config', 'action', 'cli_config',
                        'cli_display', 'save', 'rollback']:
            config = etree.Element('top')
            prev_len = len(self.device.staged)
            result = self.device.stage_config(config, cfg_type)

            self.assertEqual(len(self.device.staged), prev_len + 1)
            self.assertEqual(result, True)

    def test_stage_config_error(self):
        config = etree.Element('top')
        with self.assertRaises(ValueError):
            self.device.stage_config(config, 'bad_cfg_type')

    def test_staged_to_string(self):
        config = etree.Element('top')
        self.device.stage_config(config, 'edit_config')
        self.device.stage_config('display run', 'cli_display')

        result = self.device.staged_to_string()
        self.assertEqual(result, [b'<top/>', 'display run'])

    def test_execute_connection_closed(self):
        self.device.connection.connected = False
        with self.assertRaises(ConnectionClosedError):
            self.device.execute(self.device.connection.cli_display)

    def test_execute_rpc_error(self):
        self.device.connection.cli_display.side_effect = RPCError(etree.Element('error'))
        with self.assertRaises(NCError):
            self.device.execute(self.device.connection.cli_display, ['display current'])

    def test_execute_timeout_error(self):
        self.device.connection.cli_display.side_effect = NcOpErrors.TimeoutExpiredError
        with self.assertRaises(NCTimeoutError):
            self.device.execute(self.device.connection.cli_display, ['display current'])

    def test_execute_transport_error(self):
        self.device.connection.cli_display.side_effect = NcTransErrors.TransportError
        with self.assertRaises(ConnectionClosedError):
            self.device.execute(self.device.connection.cli_display, ['display current'])

    @mock.patch.object(HPCOM7, 'lock')
    @mock.patch.object(HPCOM7, 'unlock')
    def test_execute(self, mock_unlock, mock_lock):
        result = self.device.execute(self.device.connection.cli_display, ['display current'], {'b': 2})

        mock_lock.assert_called_with()
        mock_unlock.assert_called_with()
        self.device.connection.cli_display.assert_called_with('display current', b=2)
        self.assertEqual(result, self.device.connection.cli_display.return_value)

    @mock.patch.object(HPCOM7, 'edit_config')
    @mock.patch.object(HPCOM7, 'action')
    @mock.patch.object(HPCOM7, 'cli_config')
    @mock.patch.object(HPCOM7, 'cli_display')
    @mock.patch.object(HPCOM7, 'save')
    @mock.patch.object(HPCOM7, 'rollback')
    def test_execute_staged(self,
                            mock_rollback,
                            mock_save,
                            mock_cli_display,
                            mock_cli_config,
                            mock_action,
                            mock_edit_config):

        for cfg_type in ['edit_config', 'action', 'cli_config', 'cli_display', 'save', 'rollback']:
            self.device.staged.append(dict(cfg_type=cfg_type, config='config object'))

        self.device.execute_staged()

        for func_tuple in [
            (mock_edit_config, [], dict(target='running', config='config object')),
            (mock_action, ['config object'], {}),
            (mock_cli_config, ['config object'], {}),
            (mock_cli_display, ['config object'], {}),
            (mock_save, ['config object'], {}),
            (mock_rollback, ['config object'], {})
        ]:
            func = func_tuple[0]
            args = func_tuple[1]
            kwargs = func_tuple[2]

            func.assert_any_call(*args, **kwargs)

        self.assertEqual(len(self.device.staged), 0)

    def test_lock(self):
        self.device._locked = False

        self.device.lock()

        self.device.connection.lock.assert_called_with('running')
        self.assertEqual(self.device._locked, True)

    def test_lock_conflict_error(self):
        error = RPCError(etree.Element('lock-denied'))
        error._tag = 'lock-denied'

        self.device.connection.lock.side_effect = error
        with self.assertRaises(LockConflictError):
            self.device.lock()

    def test_lock_error(self):
        self.device.connection.lock.side_effect = RPCError(etree.Element('error'))
        with self.assertRaises(NCError):
            self.device.lock()


    def test_unlock(self):
        self.device._locked = True

        self.device.unlock()

        self.device.connection.unlock.assert_called_with('running')
        self.assertEqual(self.device._locked, False)

    def test_unlock_conflict_error(self):
        self.device._locked = True

        error = RPCError(etree.Element('operation-failed'))
        error._tag = 'operation-failed'
        error._message = 'Unlock Failed'

        self.device.connection.unlock.side_effect = error
        with self.assertRaises(UnlockConflictError):
            self.device.unlock()

    def test_unlock_error(self):
        self.device._locked = True
        self.device.connection.unlock.side_effect = RPCError(etree.Element('error'))
        with self.assertRaises(NCError):
            self.device.unlock()

    @mock.patch.object(HPCOM7, 'execute')
    def test_edit_config(self, mock_execute):
        config = etree.Element('top')
        result = self.device.edit_config(config)
        expected_kwargs = dict(config=config, target='running')

        mock_execute.assert_called_with(self.device.connection.edit_config, kwargs=expected_kwargs)
        self.assertEqual(result, mock_execute.return_value)

    @mock.patch.object(HPCOM7, 'execute')
    def test_get(self, mock_execute):
        get_tuple = ('subtree', etree.Element('top'))
        result = self.device.get(get_tuple=get_tuple)
        expected_args = [get_tuple]

        mock_execute.assert_called_with(self.device.connection.get, expected_args)
        self.assertEqual(result, mock_execute.return_value)

    @mock.patch.object(HPCOM7, 'execute')
    def test_action(self, mock_execute):
        element = etree.Element('top')
        result = self.device.action(element)
        expected_args = [element]

        mock_execute.assert_called_with(self.device.connection.action, expected_args)
        self.assertEqual(result, mock_execute.return_value)

    @mock.patch.object(HPCOM7, 'execute')
    def test_save(self, mock_execute):
        filename = 'safety.cfg'
        result = self.device.save(filename=filename)
        expected_args = [filename]

        mock_execute.assert_called_with(self.device.connection.save, expected_args)
        self.assertEqual(result, mock_execute.return_value)

    @mock.patch.object(HPCOM7, 'execute')
    def test_rollback(self, mock_execute):
        filename = 'safety.cfg'
        result = self.device.rollback(filename=filename)
        expected_args = [filename]

        mock_execute.assert_called_with(self.device.connection.rollback, expected_args)
        self.assertEqual(result, mock_execute.return_value)

    @mock.patch.object(HPCOM7, 'execute')
    def test_cli_display(self, mock_execute):
        mock_execute.return_value = RPCReply("""<?xml version="1.0" encoding="UTF-8"?><rpc-reply xmlns:config="http://www.hp.com/netconf/config:1.0" xmlns:data="http://www.hp.com/netconf/data:1.0" xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="urn:uuid:44406142-e989-11e5-a4a3-60f81db7542c"><CLI><Execution><![CDATA[<HP1>display arp
  Type: S-Static   D-Dynamic   O-Openflow   R-Rule   M-Multiport  I-Invalid
IP address      MAC address    VLAN     Interface                Aging Type
10.1.100.1      0014-1c57-a4c2 N/A      M-GE0/0/0                18    D
]]></Execution></CLI></rpc-reply>""")

        command = 'display arp'
        result = self.device.cli_display(command)
        expected = '<HP1>display arp\n  Type: S-Static   D-Dynamic   O-Openflow   R-Rule   M-Multiport  I-Invalid\nIP address      MAC address    VLAN     Interface                Aging Type\n10.1.100.1      0014-1c57-a4c2 N/A      M-GE0/0/0                18    D\n'

        mock_execute.assert_called_with(self.device.connection.cli_display, [command])
        self.assertEqual(result, expected)

    @mock.patch.object(HPCOM7, 'execute')
    def test_cli_config(self, mock_execute):
        mock_execute.return_value = RPCReply("""<?xml version="1.0" encoding="UTF-8"?><rpc-reply xmlns:config="http://www.hp.com/netconf/config:1.0" xmlns:data="http://www.hp.com/netconf/data:1.0" xmlns="urn:ietf:params:xml:ns:netconf:base:1.0" message-id="urn:uuid:5dcf4d19-e98b-11e5-8af9-60f81db7542c"><CLI><Configuration><![CDATA[<HP1>system
System View: return to User View with Ctrl+Z.
[HP1]interface FortyGigE1/0/1
]]></Configuration></CLI></rpc-reply>""")

        command = 'FortyGigE1/0/1'
        result = self.device.cli_config(command)
        expected = '<HP1>system\nSystem View: return to User View with Ctrl+Z.\n[HP1]interface FortyGigE1/0/1\n'

        mock_execute.assert_called_with(self.device.connection.cli_config, [command])
        self.assertEqual(result, expected)

    def test_reboot(self):
        self.device.reboot()
        self.device.connection.cli_display.assert_called_with(['reboot force'])

    def test_reboot_on_timeout(self):
        self.device.connection.cli_display.side_effect = NCTimeoutError

        self.device.reboot()
        self.device.connection.cli_display.assert_called_with(['reboot force'])






















if __name__ == "__main__":
    unittest.main()



"""Manage portchannels on HPCOM7 devices.
"""
from pyhpecw7.features.errors import InvalidPortType, AggregationGroupError
from pyhpecw7.utils.xml.lib import *
from pyhpecw7.features.interface import Interface
import base64
import binascii


class Portchannel(object):
    """This class is used to collect data or configure a specific portchannel.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        groupid (str): group # of the RAGG/BAGG interface
        pc_type (str): must be "bridged" or "routed"


    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        groupid (str): group # of the RAGG/BAGG interface
        pc_type (str): must be "bridged" or "routed"
        members_to_remove (list): interface names to remove from the
            portchannel
        desired_lacp_mode (str): set to "active" or "passive".  Should be
            set when ``mode`` is set to the value of ``lacp``
            in ``build_config``.

    """
    def __init__(self, device, groupid, pc_type):
        self.device = device
        self.groupid = groupid
        self.pc_type = pc_type

        # maps to internal integer representing the LAGG
        # since user facing #'s can be equal, i.e. R-Agg1
        # and B-Agg1.  This is the unique value.
        self._xgroupid = self._pc_group_mapping()
        self.members_to_remove = []
        self.desired_lacp_mode = ''

        self._members_map_index_key = {}
        self._members_map_interface_key = {}
        self._members_groups = {}

        # list of XML tags used to build proper objects
        self.pc_tags = ['LAGG', 'LAGGGroups', 'LAGGGroup']
        self.member_tags = ['LAGG', 'LAGGMembers', 'LAGGMember']

        # The rest of the attributes listed below are all used to map
        # XML tags to dictionary keys (and vice versa) as well as
        # map values coming from the switch into more user friendly values
        self.PORTCHANNEL = dict(
            groupid='GroupId',
            mode='LinkMode',
            pc_index='IfIndex',
            members='MemberList',
            lacp_edge='LacpEdgeEnable'
            )
        self.R_PORTCHANNEL = dict(reversed(
            item) for item in self.PORTCHANNEL.items())
        self.value_map = {
            'LinkMode': {
                '1': 'static',
                '2': 'dynamic'
                },
            'LacpEdgeEnable': {
                'true': 'enabled',
                'false': 'disabled'
                }
        }
        self.R_value_map = reverse_value_map(
            self.R_PORTCHANNEL, self.value_map)
        self.LACP = dict(
            groupid='GroupId',
            intf_index='IfIndex',
            enabled='LacpEnable',
            lacp_mode='LacpMode',
            )
        self.R_LACP = dict(reversed(item) for item in self.LACP.items())
        self.lacp_value_map = {
            'LacpMode': {
                '1': 'active',
                '2': 'passive'
            }
        }
        self.R_lacp_value_map = reverse_value_map(
            self.R_LACP, self.lacp_value_map)

    def get_portchannels(self):
        """Get a list of portchannel groups that exist on the switch

            Returns:
                This returns a list of numbers represented as strings
                that are the portchannel groups that exist on the switch.
        """
        E = data_element_maker()
        top = E.top(
            E.LAGG(
                E.LAGGGroups(
                    E.LAGGGroup(
                        E.IfIndex(),
                        E.GroupId(),
                        E.LinkMode(),
                        E.MemberList(),
                        E.LacpEdgeEnable()
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        pc_groups_xml = findall_in_data('GroupId', nc_get_reply.data_ele)

        groups = [group.text for group in pc_groups_xml]

        return groups

    def get_config(self):
        """Get current configuration for a given portchannel

        Returns:
            This returns a dictionary that has the following k/v pairs
            if the portchannel exists:

                :groupid (str): group ID of the portchannel
                :ncgroupid (str): INTERNAL group ID used by the switch
                    to differentiate between bridged, routed, and other types
                    of LAGGs. Kept to assist in troubleshooting.
                :mode (str): will be "static" or "dynamic"
                :members (list): list of current members by interface name
                :min_ports (str): number that represents selected-port minimum
                :max_ports (str): number that represents selected-port maximum
                :lacp_modes_by_interface (dict): list of dicts that have two
                    key/value pairs. sample_dict=(interface='FortyGigE1/0/1',
                    lacp_mode='passive')

            It returns an empty dictionary if the portchannel group does
            not exist.

        """
        E = data_element_maker()
        top = E.top(
            E.LAGG(
                E.LAGGGroups(
                    E.LAGGGroup(
                        E.GroupId(self._xgroupid)
                    )
                )
            )
        )
        nc_get_reply = self.device.get(('subtree', top))
        return_pc = data_elem_to_dict(nc_get_reply.data_ele, self.PORTCHANNEL, value_map=self.value_map)

        if return_pc:
            return_pc['groupid'] = self.groupid
            return_pc['nc_groupid'] = self._xgroupid

            members = []
            members_by_index = []
            members_by_name = []
            if return_pc.get('members'):
                members_by_index, members_by_name = \
                    self._get_members_from_bitmap(return_pc.get('members'))

                # building dictionary that has keys that are IfIndex values
                # and has values that are the associated interface name
                for count, each in enumerate(members_by_index):
                    self._members_map_index_key[str(each)] = \
                        members_by_name[count]

                # building dictionary that has keys that are Interface names
                # and has values that are the associated IfIndex values
                for count, each in enumerate(members_by_name):
                    self._members_map_interface_key[each.lower()] = \
                        members_by_index[count]

                for memb in members_by_name:
                    mode = self.get_lacp_mode_by_name(name=memb)
                    temp = dict(interface=memb, lacp_mode=mode)
                    members.append(temp)

            return_pc['members'] = members_by_name
            return_pc['lacp_modes_by_interface'] = members

            self._get_pc_config_raw()

            return_pc['min_ports'] = self.get_selected_port_min()
            return_pc['max_ports'] = self.get_selected_port_max()

        return return_pc

    def get_all_members(self, list_type='name', asdict=False):
        """Gets ports that are a member to any port channel

           Args:
               list_type (str): must be "name" or "ifindex"
               asdict (bool): determines if a dict should be returned
                   this overrides list_type (see Returns)

           Returns:
               1 of 4 objects can be returned based on input args:

                   if asdict=True regardless of other Args, a dict
                       is returned that has interface names as keys and
                       the group of the port-channel that is config'd on that
                       interface as the key
                    if list_type is name (default), a list of interface
                        names is returned.  The names of interfaces that
                        have any portchannel config'd.
                    if list_type is set to "ifindex", the list has all
                        ifindexes instead of names of interfaces that have
                        a portchannel config'd.
                    if list_type is misconfigured, an error string is returned.
        """
        E = data_element_maker()
        top = E.top(
            E.LAGG(
                E.LAGGMembers(
                    E.LAGGMember(
                        E.IfIndex(),
                        E.GroupId()
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        data = findall_in_data('LAGGMember', nc_get_reply.data_ele)

        members_as_index = []
        members_as_name = []

        for each in data:
            index = find_in_data('IfIndex', each).text
            group = find_in_data('GroupId', each).text
            if group != '0':
                members_as_index.append(index)
                name = self.get_interface_from_index(index)
                members_as_name.append(name)
                self._members_groups[name] = group

        if asdict:
            return self._members_groups
        if list_type == 'name':
            return members_as_name
        elif list_type == 'ifindex':
            return members_as_index
        else:
            return 'invalid value for list_type'

    def get_lacp_mode_by_name(self, name):
        """Get current LACP mode for a given interface

           Args:
               name (str): full name of the interface

           Returns:
               mode (str): "active" or "passive"
        """
        E = data_element_maker()
        top = E.top(
            E.LAGG(
                E.LAGGMembers(
                    E.LAGGMember(
                        E.IfIndex(self.get_index_from_interface(name))
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        return_lacp = data_elem_to_dict(nc_get_reply.data_ele, self.LACP, value_map=self.lacp_value_map)

        # if return_lacp.get('groupid') != '0':
        #    return_lacp['groupid'] = self.groupid

        return return_lacp.get('lacp_mode')

    def _get_members_from_bitmap(self, bitmap):
        """Return list of interface names from bitmap encoded as base64

            Args:
                bitmap (str): memberlist as base64 as retrieved via NETCONF

            Returns:
                This returns a list of interface names.

        """

        decoded = base64.b64decode(bitmap)
        if isinstance(decoded, str):
            hex_value = decoded.encode('hex')
        else:
            hex_value = decoded.hex()
        h_size = len(hex_value) * 4
        binary_value = (bin(int(hex_value, 16))[2:]).zfill(h_size)
        members_by_index = []
        for index, bit in enumerate(binary_value):
            port = index + 1
            if bit == '1':
                members_by_index.append(port)
        members_by_name = []
        for index in members_by_index:
            members_by_name.append(self.get_interface_from_index(str(index)))

        return members_by_index, members_by_name

#    def _get_bitmap_from_members(self, members):
#        """Return bitmap given a list of interface names
#           Not used, but kept, just in case :)
#        """
#        index_list = []
#        for member in members:
#            index = self.get_index_from_interface(member)
#            index_list.append(int(index))
#
#        index_list.sort()
#        binary = ''
#        start = 1
#        for finish in index_list:
#            for num in range(start, finish):
#                binary += '0'
#            binary += '1'
#            start = finish + 1
#
#        lenbinary = len(binary)
#        start = 8
#        while (start % lenbinary) == start:
#            start += 8
#        newfill = start - len(binary)
#        binary += newfill * '0'
#
#        # chr only goes up to 255!!! fixed in py3!!! ugh
#        # bitmap = base64.b64encode(chr(int(binary, 2)))
#        hexa = hex(int(binary, 2))[2:]
#        if len(hexa) % 2 != 0:
#            hexa += '0'
#        converted_to_ascii = binascii.a2b_hex(hexa)
#        bitmap = base64.b64encode(converted_to_ascii)
#
#        return bitmap

    def get_interface_from_index(self, index):
        """Return interface name based on a given ifindex
        """
        E = data_element_maker()
        top = E.top(
            E.Ifmgr(
                E.Interfaces(
                    E.Interface(
                        E.IfIndex(index)
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        interface_name = find_in_data('Name', nc_get_reply.data_ele).text

        return interface_name

    def _pc_group_mapping(self):
        """Map user input for portchannel group to the internal integer
           used by the system, since the user can create a routed-agg with
           group 1 and bridged-group with group 1. This maps the group to the
           "real" group used internally
        """

        group = int(self.groupid)
        if self.pc_type == 'bridged':
            self._xgroupid = group
        elif self.pc_type == 'routed':
            self._xgroupid = group + 16384
        return str(self._xgroupid)

    def _get_pc_config_raw(self):
        """Get raw text from CLI current configuration for the portchannel
        """
        if self.pc_type == 'bridged':
            self.fulltype = 'Bridge-Aggregation'
        elif self.pc_type == 'routed':
            self.fulltype = 'Route-Aggregation'

        text = self.device.cli_display(
            'display current-configuration interface {0} {1}'.format(
                self.fulltype, self.groupid))

        text_as_list = text.split('\n')

        self.raw_config = text_as_list

    def get_selected_port_min(self):
        """Get selected port min configuration

            Returns:
                This returns the selected-port minimum
                configured value on the switch, else it returns None

        """
        if not self.raw_config:
            self._get_pc_config_raw()

        find = 'selected-port minimum'
        for each in self.raw_config:
            if find in each:
                min_value = each.split(find)[-1].strip()
                return min_value
        return None

    def get_selected_port_max(self):
        """Get selected port max configuration

            Returns:
                This returns the selected-port maximum
                configured value on the switch, else it returns None
        """
        if not self.raw_config:
            self._get_pc_config_raw()

        find = 'selected-port maximum'
        for each in self.raw_config:
            if find in each:
                max_value = each.split(find)[-1].strip()
                return max_value
        return None

    def _get_min_max_cmds(self, **portchannel):
        """Get commands required to config min/max ports

            Args:
                portchannel: see Keyword Args

            Keyword Args:
                min_ports: value to set for selected-port minimum
                max_ports: value to set for selected-port maximum

        """
        min_links = portchannel.get('min_ports')
        max_links = portchannel.get('max_ports')
        commands = []
        cmd = 'link-aggregation selected-port'
        if min_links:
            command = cmd + ' minimum {0}'.format(min_links)
            if command:
                commands.append(command)
        if max_links:
            command = cmd + ' maximum {0}'.format(max_links)
            if command:
                commands.append(command)
        if min_links or max_links:
            commands.insert(0, 'interface {0} {1}'.format(self.fulltype,
                                                          self.groupid))

        return commands

    def get_index_from_interface(self, interface):
        """Get IfIndex from interface name

            Args:
                interface (str): name of the interface

            Returns:
                This returns the IfIndex for an interface.

        """

        local_index = self._members_map_interface_key.get(interface)
        if local_index:
            index = local_index
        else:
            intf = Interface(self.device, interface)
            index = str(intf.iface_index)
            self._members_map_interface_key[interface] = index

        return index

    def _add_lagg_member(self, interface, remove=False, lacp=None):
        """Add lagg member to an XML object element
        """
        if remove:
            group = '0'
        else:
            group = self._xgroupid

        E = config_element_maker()

        lacp_args = []
        if lacp:
            lacp_args.append(E.LacpMode(lacp))

        member = E.LAGGMember(
            E.IfIndex(self.get_index_from_interface(interface)),
            E.GroupId(group),
            *lacp_args
        )

        return member

    def remove(self, stage=False):
        """Stage or execute a config object to remove portchannel

        Args:
            stage (bool): whether to stage the commands or execute
                immediately

        Returns:
            True if stage=True and successfully staged
            List of etree.Element XML responses if immediate execution
        """
        return self._build_config(state='absent', stage=stage)

    def build(self, stage=False, **portchannel):
        """Stage or execute a config object to add/update portchannel

        Args:
            state (str): must be "present" or "absent"
            portchannel: see Keyword Args
            stage (bool): whether to stage the commands or execute
                immediately

        Keyword Args:
            members (list): OPTIONAL - list of members by interface name
                being configured
            min_ports (str): OPTIONAL - number that represents
                selected-port minimum
            max_ports (str): OPTIONAL - number that represents
                selected-port maximum
            lacp_to_change (list): OPTIONAL - list of interfaces that need
                have their lacp mode changed

        Note:
            ``desired_lacp_mode`` needs to be set for the members in
            ``portchannel['lacp_to_change']`` to take effect.

            ``members_to_remove`` can be set to remove members during the
            build process.  This should also be a list of interface names.

        Returns:
            True if stage=True and successfully staged
            List of etree.Element XML responses if immediate execution
        """
        return self._build_config(state='present', stage=stage, **portchannel)

    def _build_config(self, state, stage=False, **portchannel):
        """Stage or execute a config object to add/update portchannel

        Args:
            stage (bool): whether to stage the commands or execute
                immediately

        Returns:
            True if stage=True and successfully staged
            List of etree.Element XML responses if immediate execution
        """
        if state == 'present':
            operation = 'merge'
        elif state == 'absent':
            operation = 'delete'

        # needs to be the "internal" id, not user exposed id
        portchannel['groupid'] = self._xgroupid

        EC = nc_element_maker()
        E = config_element_maker()

        members_desired_list = portchannel.get('members') or []

        # removing members b/c it messes with the value map
        if 'members' in portchannel.keys():
            portchannel.pop('members')

        config = EC.config(
            E.top(
                E.LAGG(
                    E.LAGGGroups(
                        E.LAGGGroup(
                            *config_params(portchannel, self.PORTCHANNEL, value_map=self.R_value_map, fill_in=False)
                        )
                    )
                ),
                **operation_kwarg(operation)
            )
        )

        lacp_to_change = portchannel.get('lacp_to_change')

        if state == 'present':
            if members_desired_list or self.members_to_remove or \
                    lacp_to_change:
                lagg = find_in_config('LAGG', config)
                members = E.LAGGMembers()
                # member = etree.Element(qualify('LAGGMember', HPCONFIG))
                # if members_as_passed_in:
                for each in members_desired_list:
                    members.append(self._add_lagg_member(each))
                # now remove any members not in the desired members list
                # and set in the remove list
                for each in self.members_to_remove:
                    members.append(self._add_lagg_member(each, remove=True))
                if lacp_to_change:
                    for each in lacp_to_change:
                        value = self.R_lacp_value_map.get('lacp_mode').get(
                            self.desired_lacp_mode)
                        member = self._add_lagg_member(each, lacp=value)
                        members.append(member)

                lagg.append(members)

        # stages the native NETCONF XML Objects
        c1 = True
        c2 = True
        if stage:
            c1 = self.device.stage_config(config, 'edit_config')
        else:
            c1 = self.device.edit_config(config)

        # configuring the min and max ports not possible via NETCONF
        # Use CLI via NETCONF instead, separate staging happening here too
        if portchannel.get('min_ports') or portchannel.get('max_ports'):
            self._get_pc_config_raw()
            min_max = self._get_min_max_cmds(**portchannel)
            if min_max:
                if stage:
                    c2 = self.device.stage_config(min_max, 'cli_config')
                else:
                    c2 = self.device.cli_config(min_max)

        if stage:
            return c1 and c2
        else:
            return [c1, c2]

    def param_check(self, **portchannel):
        """Param validation for portchannel

            Args:
                state (str): present or absent
                portchannel: see Keyword Args

            Keyword Args:
                members (list): members by interface name being configured

            Raises:
                InvalidPortType: when existing port type does not match
                    desired type of portchannel
                AggregationGroupError: when an interface is already
                    a member of a different portchannel than the
                    one being configured.

        """
        if portchannel.get('members'):
            members = portchannel.get('members')
            for each in members:
                interf = Interface(self.device, each)
                configured_type = interf.get_config().get('type')
#                if configured_type == 'bridged':
#                    configured_type = 'bridged'
                if configured_type != self.pc_type:
                    raise InvalidPortType(each, configured_type,
                                          self.pc_type)
                member_dict = self.get_all_members(asdict=True)
                existing_group = member_dict.get(each)
                if existing_group:
                    if existing_group != self._xgroupid:
                        raise AggregationGroupError(each)



import unittest
import mock
from lxml import etree

from ncclient.operations.retrieve import GetReply
from pyhpecw7.features.vlan import Vlan
from pyhpecw7.features.errors import VlanIDError, LengthOfStringError

from .base_feature_test import BaseFeatureCase

class VlanTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.vlan = Vlan(self.device, vlanid='77')

    def test_get_vlan_list(self):
        expected_get, get_reply = self.xml_get_and_reply('vlan_list')
        self.device.get.return_value = get_reply

        expected = ['1', '20', '77']

        vlan_list = self.vlan.get_vlan_list()

        self.assertEqual(vlan_list, expected)
        self.assert_get_request(expected_get)

    def test_get_config(self):
        expected_get, get_reply = self.xml_get_and_reply('vlan')
        self.device.get.return_value = get_reply

        expected = {'name': 'hello', 'vlanid': '77', 'descr': 'goodbye'}

        vlan = self.vlan.get_config()

        self.assertEqual(vlan, expected)
        self.assert_get_request(expected_get)

    def test_build_config(self):
        result = self.vlan._build_config(state='present')
        expected = self.read_config_xml('vlan')
        self.assert_elements_equal(result, expected)

        result = self.vlan._build_config(state='present', name='hello')
        expected = self.read_config_xml('vlan_name')
        self.assert_elements_equal(result, expected)

        result = self.vlan._build_config('present', name='hello', descr='goodbye')
        expected = self.read_config_xml('vlan_name_descr')
        self.assert_elements_equal(result, expected)

        result = self.vlan._build_config('absent')
        expected = self.read_config_xml('vlan_absent')
        self.assert_elements_equal(result, expected)

    def test_param_check(self):
        with self.assertRaises(LengthOfStringError):
            self.vlan.param_check(name=('a' * 255))

        with self.assertRaises(LengthOfStringError):
            self.vlan.param_check(descr=('b' * 255))

    @mock.patch.object(Vlan, '_build_config')
    def test_build(self, mock_build_config):
        self.vlan.build(name='a', descr='b')
        mock_build_config.assert_called_with(state='present', name='a', descr='b')
        self.device.edit_config.assert_called_with(mock_build_config.return_value)

        self.vlan.build(stage=False, name='a', descr='b')
        mock_build_config.assert_called_with(state='present', name='a', descr='b')
        self.device.edit_config.assert_called_with(mock_build_config.return_value)

        self.vlan.build(stage=True, name='a', descr='b')
        mock_build_config.assert_called_with(state='present', name='a', descr='b')
        self.device.stage_config.assert_called_with(mock_build_config.return_value, 'edit_config')

    @mock.patch.object(Vlan, '_build_config')
    def test_remove(self, mock_build_config):
        self.vlan.remove()
        mock_build_config.assert_called_with(state='absent')
        self.device.edit_config.assert_called_with(mock_build_config.return_value)

        self.vlan.remove(stage=False)
        mock_build_config.assert_called_with(state='absent')
        self.device.edit_config.assert_called_with(mock_build_config.return_value)

        self.vlan.remove(stage=True)
        mock_build_config.assert_called_with(state='absent')
        self.device.stage_config.assert_called_with(mock_build_config.return_value, 'edit_config')


if __name__ == '__main__':
    unittest.main()


import unittest
import mock
import os

from pyhpecw7.features.config import Config
from .base_feature_test import BaseFeatureCase

FILENAME = '/path/to/file.cfg'
BASENAME = os.path.basename(FILENAME)

class ConfigTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.config = Config(self.device, FILENAME)

    def test_init(self):
        self.assertEqual(self.config.filename, FILENAME)
        self.assertEqual(self.config.basename, BASENAME)

    def test_get_diffs(self):
        result = self.config._get__diffs_from_switch()
        self.device.cli_display.assert_called_with('display diff current-configuration configfile file.cfg')

    def test_compare_config(self):
        self.config._diff_response = self.read_cli_display('display_diff')

        result = self.config.compare_config()
        expected = (['+port group interface FortyGigE1/0/1', '+port group interface FortyGigE1/0/2', '+irf domain 1', '+flash:/safety_file.cfg', '+name VLAN_77', '+description instant vlan', '+name web_vlan', '+irf-port 1/2', '+irf-port 1/1', '-name VLAN100_VRRP_TEST', '-description vlan 3 for testing', '-destination 10.1.1.2', '-mad exclude interface FortyGigE1/0/10', '-interface LoopBack29', '-interface Tunnel22 mode vxlan', '-vrrp vrid 100 authentication-mode md5 cipher $c$3$sBCWj6+b6pWNZe7D6EQSfsG4j4ca1SASow==', '-interface Vlan-interface100', '-lacp mode passive', '-vrrp vrid 100 priority 10', '-name hello', '-name VLAN5_TEST', '-vlan 100', '-Current configuration', '-interface Bridge-Aggregation3', '-port link-mode route', '-mad exclude interface FortyGigE1/0/9', '-vlan 3', '-ip address 100.100.100.2 255.255.255.0', '-vrrp vrid 100 virtual-ip 100.100.100.1', '-source 10.1.1.1', '-interface Tunnel20 mode vxlan', '-interface Tunnel24 mode vxlan', '-irf domain 2', '-tunnel global source-address 10.10.10.10', '-vlan 2', '-description goodbye', '-description vlan 5 for testing', '-interface Tunnel21 mode vxlan', '-name VLAN3_TEST', '-interface Tunnel23 mode vxlan', '-l2vpn enable'], ['<HP1>display diff current-configuration configfile safety_file.cfg', '--- Current configuration', '+++ flash:/safety_file.cfg', '@@ -5,7 +5,7 @@', ' #', '  clock protocol none', ' #', '- irf domain 2', '+ irf domain 1', '  irf mac-address persistent timer', '  irf auto-update enable', '  undo irf link-delay', '@@ -13,8 +13,6 @@', '  irf member 1 description They used to call me HP1', '  irf mode normal', ' #', '- tunnel global source-address 10.10.10.10', '-#', '  lldp global enable', ' #', '  system-working-mode standard', '@@ -23,72 +21,31 @@', ' #', ' vlan 1', ' #', '-vlan 2', '-#', '-vlan 3', '- name VLAN3_TEST', '- description vlan 3 for testing', '-#', ' vlan 5', '- name VLAN5_TEST', '- description vlan 5 for testing', '+ name web_vlan', ' #', ' vlan 20', '  name VLAN20', ' #', ' vlan 77', '- name hello', '- description goodbye', '+ name VLAN_77', '+ description instant vlan', ' #', '-vlan 100', '- name VLAN100_VRRP_TEST', '+irf-port 1/1', '+ port group interface FortyGigE1/0/1', ' #', '+irf-port 1/2', '+ port group interface FortyGigE1/0/2', '+#', '  stp global enable', ' #', '- l2vpn enable', '-#', '-interface Bridge-Aggregation3', '-#', ' interface NULL0', ' #', '-interface LoopBack29', '-#', '-interface Vlan-interface100', '- ip address 100.100.100.2 255.255.255.0', '- vrrp vrid 100 virtual-ip 100.100.100.1', '- vrrp vrid 100 authentication-mode md5 cipher $c$3$sBCWj6+b6pWNZe7D6EQSfsG4j4ca1SASow==', '- vrrp vrid 100 priority 10', '-#', '-interface FortyGigE1/0/9', '- port link-mode route', '-#', '-interface FortyGigE1/0/27', '- port link-mode route', '-#', '-interface FortyGigE1/0/28', '- port link-mode route', '-#', '-interface FortyGigE1/0/29', '- port link-mode route', '-#', '-interface FortyGigE1/0/30', '- port link-mode route', '-#', '-interface FortyGigE1/0/1', '- port link-mode bridge', '- shutdown', '-#', '-interface FortyGigE1/0/2', '- port link-mode bridge', '- shutdown', '-#', ' interface FortyGigE1/0/3', '  port link-mode bridge', '- shutdown', ' #', ' interface FortyGigE1/0/4', '  port link-mode bridge', '- shutdown', ' #', ' interface FortyGigE1/0/5', '  port link-mode bridge', '@@ -102,12 +59,14 @@', ' interface FortyGigE1/0/8', '  port link-mode bridge', ' #', '+interface FortyGigE1/0/9', '+ port link-mode bridge', '+#', ' interface FortyGigE1/0/10', '  port link-mode bridge', ' #', ' interface FortyGigE1/0/11', '  port link-mode bridge', '- lacp mode passive', ' #', ' interface FortyGigE1/0/12', '  port link-mode bridge', '@@ -154,35 +113,32 @@', ' interface FortyGigE1/0/26', '  port link-mode bridge', ' #', '-interface FortyGigE1/0/31', '+interface FortyGigE1/0/27', '  port link-mode bridge', ' #', '-interface FortyGigE1/0/32', '+interface FortyGigE1/0/28', '  port link-mode bridge', ' #', '-interface M-GigabitEthernet0/0/0', '- ip address 10.1.100.40 255.255.255.0', '+interface FortyGigE1/0/29', '+ port link-mode bridge', ' #', '-interface Tunnel20 mode vxlan', '- source 10.1.1.1', '- destination 10.1.1.2', '+interface FortyGigE1/0/30', '+ port link-mode bridge', ' #', '-interface Tunnel21 mode vxlan', '- source 10.1.1.1', '- destination 10.1.1.2', '+interface FortyGigE1/0/31', '+ port link-mode bridge', ' #', '-interface Tunnel22 mode vxlan', '- source 10.1.1.1', '- destination 10.1.1.2', '+interface FortyGigE1/0/32', '+ port link-mode bridge', ' #', '-interface Tunnel23 mode vxlan', '- source 10.1.1.1', '- destination 10.1.1.2', '+interface FortyGigE1/0/1', '+ shutdown', ' #', '-interface Tunnel24 mode vxlan', '- source 10.1.1.1', '- destination 10.1.1.2', '+interface FortyGigE1/0/2', ' #', '+interface M-GigabitEthernet0/0/0', '+ ip address 10.1.100.40 255.255.255.0', '+#', '  scheduler logfile size 16', ' #', ' line class aux', '@@ -205,9 +161,6 @@', '  user-role network-operator', ' #', '  ip route-static 0.0.0.0 0 10.1.100.1', '-#', '- mad exclude interface FortyGigE1/0/9', '- mad exclude interface FortyGigE1/0/10', ' #', '  ssh server enable', '  ssh user hp service-type all authentication-type password'])

        self.assertEqual(set(result[0]), set(expected[0]))
        self.assertEqual(set(result[1]), set(expected[1]))

    def test_build(self):
        self.config.build()
        self.device.save.assert_any_call('safety_file.cfg')
        self.device.rollback.assert_called_with(self.config.basename)
        self.device.save.assert_any_call('startup.cfg')

        self.config.build(stage=True)
        self.assert_stage_requests_multiple([['safety_file.cfg', 'save'], [self.config.basename, 'rollback'], ['startup.cfg', 'save']])




if __name__ == "__main__":
    unittest.main()


import unittest
import mock

from pyhpecw7.features.cleanerase import CleanErase
from .base_feature_test import BaseFeatureCase

class CleanEraseTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.clean_erase = CleanErase(self.device)

    def test_build(self):
        self.clean_erase.build(factory_default=True)
        self.device.cli_display.assert_called_with(['restore factory-default'])

        self.clean_erase.build(factory_default=True, stage=True)
        self.device.stage_config.assert_called_with(['restore factory-default'], 'cli_display')
        self.device.reboot.assert_called_with()

        result = self.clean_erase.build()
        self.assertEqual(result, None)

if __name__ == "__main__":
    unittest.main()


import unittest
import mock

from pyhpecw7.features.neighbor import Neighbors
from .base_feature_test import BaseFeatureCase


class NeighborTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.neighbors = Neighbors(self.device)

    @mock.patch.object(Neighbors, '_get_interface_from_index')
    def test_neighbors_lldp(self, mock_index):
        mock_index.return_value = 'FortyGigE1/0/10'
        expected_get, get_reply = self.xml_get_and_reply('neighbors_lldp')
        self.device.get.return_value = get_reply

        expected = [
            {
                'local_intf': 'FortyGigE1/0/10',
                'neighbor': 'PERIMETER.cisco.com'
            }
        ]

        result = self.neighbors._get_neighbors()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Neighbors, '_get_interface_from_index')
    def test_neighbors_cddp(self, mock_index):
        mock_index.return_value = 'FortyGigE1/0/10'
        expected_get, get_reply = self.xml_get_and_reply('neighbors_cdp')
        self.device.get.return_value = get_reply

        expected = []

        result = self.neighbors._get_neighbors(ntype='cdp')

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Neighbors, '_get_neighbors')
    def test_refresh(self, mock_get_neighbors):
        self.neighbors.refresh()

        mock_get_neighbors.assert_has_calls([mock.call(ntype='lldp'), mock.call(ntype='cdp')])
        self.assertEqual(self.neighbors.cdp, mock_get_neighbors.return_value)
        self.assertEqual(self.neighbors.lldp, mock_get_neighbors.return_value)


    def test_get_name_from_index(self):
        expected_get, get_reply = self.xml_get_and_reply('neighbors_iface')
        self.device.get.return_value = get_reply

        expected = 'FortyGigE1/0/3'

        iface = self.neighbors._get_interface_from_index('11')

        self.assertEqual(iface, expected)
        self.assert_get_request(expected_get)

if __name__ == '__main__':
    unittest.main()


"""Reboot HPCOM7 devices.
"""
from pyhpecw7.features.errors import RebootDateError, RebootTimeError


class Reboot(object):
    """This class is used to reboot a HP COM7 switch.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.

    """
    def __init__(self, device):
        self.device = device

    def build(self, stage=False, **reboot):
        """Build command list to reboot the switch and send to staging

        Args:
            stage (bool): whether to stage the commands or execute
                immediately
            reboot: see Keyword Args

        Keyword Args:
            reboot (bool): REQUIRED - set to True to reboot (safety)
            time (str): OPTIONAL - must be in HH:MM format
            date (str): OPTIONAL - must be in MM/DD/YYYY format
            delay (str): OPTIONAL - number representing delay in minutes

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediate execution
        """

        reb2 = reboot.get('reboot')
        time = reboot.get('time')
        delay = reboot.get('delay')
        date = reboot.get('date')

        if reb2:
            if delay:
                commands = ['scheduler reboot delay {0}'.format(delay)]
            elif time:
                if date:
                    commands = ['scheduler reboot at {0} {1}'.format(time,
                                                                     date)]
                else:
                    commands = ['scheduler reboot at {0}'.format(time)]
            else:
                commands = ['reboot force']
                # will exit/fail, NETCONF connection will not be closed

        if stage:
            return self.device.stage_config(commands, 'cli_display')
        else:
            return self.device.cli_display(commands)

    def param_check(self, **reboot):
        """Param validation for time & date.

        Args:
            reboot: see Keyword Args

        Keyword Args:
            time (str): OPTIONAL - must be in HH:MM format
            date (str): OPTIONAL - must be in MM/DD/YYYY format

        """

        time = reboot.get('time')
        if time:
            if len(time) != 5 or ':' not in time:
                raise RebootTimeError
            time_hh_mm = time.split(':')
            if len(time_hh_mm) != 2:
                raise RebootTimeError
            mm = time_hh_mm[0]
            hh = time_hh_mm[1]
            if len(mm) != 2 or len(hh) != 2:
                raise RebootTimeError

        date = reboot.get('date')
        if date:
            if '/' not in date:
                raise RebootDateError
            mm_dd_yyyy = date.split('/')
            if len(mm_dd_yyyy) != 3:
                raise RebootDateError
            mm = mm_dd_yyyy[0]
            dd = mm_dd_yyyy[1]
            yyyy = mm_dd_yyyy[2]
            if len(mm) != 1 and len(mm) != 2:
                raise RebootDateError
            if len(dd) != 1 and len(dd) != 2:
                raise RebootDateError
            if len(yyyy) != 4:
                raise RebootDateError



"""This module provides several utility functions
for dealing with XML text and ``etree.Element``
XML objects.
"""
from lxml.builder import ElementMaker

from pyhpecw7.utils.xml.namespaces import *


def config_element_maker():
    return ElementMaker(namespace=HPCONFIG, nsmap={None: HPCONFIG})


def data_element_maker():
    return ElementMaker(namespace=HPDATA, nsmap={None: HPDATA})


def action_element_maker():
    return ElementMaker(namespace=HPACTION, nsmap={None: HPACTION})


def nc_element_maker():
    return ElementMaker(namespace=NETCONFBASE, nsmap={None: NETCONFBASE})


def config_params(pmap, key_map, value_map={}, E=config_element_maker(), fill_in=True):
    params = []

    for key, value in pmap.items():
        if fill_in:
            key2 = key_map.get(key, key)
        else:
            key2 = key_map.get(key, None)

        if key2:
            params.append(
                getattr(E, key2)(value_map.get(key, {}).get(value, value)))

    return params


def operation_kwarg(operation=None):
    if operation is None:
        return {}

    return {
        NETCONFBASE_C + 'operation': operation
    }


def _findall_with_ns(query, ele, ns=''):
    return ele.findall('.//{%s}%s' % (ns, query))


def findall_in_data(query, ele):
    return _findall_with_ns(query, ele, ns=HPDATA)


def _find_with_ns(query, ele, ns=''):
    return ele.find('.//{%s}%s' % (ns, query))


def find_in_data(query, ele):
    return _find_with_ns(query, ele, ns=HPDATA)


def find_in_action(query, ele):
    return _find_with_ns(query, ele, ns=HPACTION)


def findall_in_action(query, ele):
    return _findall_with_ns(query, ele, ns=HPACTION)


def find_in_config(query, ele):
    return _find_with_ns(query, ele, ns=HPCONFIG)


def elem_to_dict(elem, ns, key_map, value_map={}):
    """Convert an XML etree.Element to a desired dictionary
    as specified by the key map and value map.
    Args:
        elem (etree.Element): An ancestor element
            of the tags specified in the key map.
        ns (string): The namespace to use
            when searching for XML tags.
        key_map (dict): A mapping from desired
            dictionary keys to XML tag names.
        value_map (dict): A mapping from XML tag names to
            dictionaries of mappings from XML text
            values to desired dictionary values.
    Returns:
        The desired dictionary.
    """
    to_dict = {}
    for k, v in key_map.items():
        field = elem.find('.//{0}{1}'.format(ns, v))
        if field is not None:
            text = field.text
            to_dict[k] = value_map.get(v, {}).get(text, text)

    return to_dict


def data_elem_to_dict(elem, key_map, value_map={}):
    return elem_to_dict(elem, HPDATA_C, key_map, value_map=value_map)


def reverse_value_map(key_map, value_map):
    """Utility function for creating a
    "reverse" value map from a given key map and value map.
    """
    r_value_map = {}
    for k, v in value_map.items():
        sub_values = r_value_map[key_map[k]] = {}
        for k2, v2 in v.items():
            sub_values[v2] = k2

    return r_value_map


def remove_namespaces(xml):
    """Remove the namespaces from an
    ``etree.Element`` object and return
    the modified object.
    """
    for elem in xml.getiterator():
        split_elem = elem.tag.split('}')
        if len(split_elem) > 1:
            elem.tag = split_elem[1]
    return xml


def get_text(xml, tag):
    """Return the text from a given tag and XML element.
    """
    elem = xml.find(tag)
    if elem is not None:
        return elem.text.strip()



"""Mange config files on HPCOM7 devices.
"""
import os
import time
from pyhpecw7.features.errors import InvalidConfigFile
from ncclient.operations.rpc import RPCError


class Config(object):
    """This class is used to activate a new running config in real-time.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        filename (str): absolute path to the file that will be compared
            and/or activated on the device.

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        filename (str): absolute path to the file that will be compared
            and/or activated on the device.

    """
    def __init__(self, device, filename):
        self.device = device
        self.filename = filename
        self.basename = os.path.basename(self.filename)
        self._diff_response = ''
        self._switch_response = []
        self._original__diffs = []
        self._diffs = []

    def _get__diffs_from_switch(self):
        """Compare switch running config to desired new config file.
        """
        cmds = 'display diff current-configuration configfile {0}'.format(
            self.basename)

        # raw string output taken directly from the switch showing the diff
        self._diff_response = self.device.cli_display(cmds)

    def compare_config(self):
        """Compare new config file to the existing current running config

        Returns:
            This returns a tuple of two elements that are both lists.

            The first element has a summary of diffs (self._diffs) and
            the second element is the exact output from the 'display diff'
            command, but as a list (self._original__diffs).
        """
        self._diffs = []
        new_cfg = []
        current_cfg = []

        if not self._diff_response:
            self._get__diffs_from_switch()

        # same as _diff_response, but as a list with each line as an element
        self._original__diffs = self._diff_response.split('\n')

        for line in self._original__diffs:
            if line.strip().startswith('-') and '#' not in line:
                current_cfg.append(line.strip('-').strip())
            elif line.strip().startswith('+') and '#' not in line:
                new_cfg.append(line.strip('+').strip())

        commands_to_apply = set(new_cfg).difference(current_cfg)
        commands_to_remove = set(current_cfg).difference(new_cfg)

        # diffs is summary of exactly which lines are changing in
        # _original__diffs because _diffs from switch show everything
        for each in commands_to_apply:
            self._diffs.append('+' + each)

        for each in commands_to_remove:
            self._diffs.append('-' + each)

        return self._diffs, self._original__diffs

#    def activate_replacement_config(self):
#        """Activate replacement config on host device
#
#        This method immediately activates the new config
#        file to be the new running config and does not perform a backup
#        or new save as is done in ``build_config``.
#
#        """
#        try:
#            # CLI response from switch during activation.
#            self._switch_response =  self.device.rollback(self.basename)
#        except RPCError as err:
#            if str(err).strip() == 'Operation failed.':
#                raise InvalidConfigFile
#
#        time.sleep(.5)

    def build(self, stage=False):
        """Stage or execute configuration required to activate new config file.

        This method stores the existing running configuration
        as flash:/safety_file.cfg, loads the new config file,
        and then stores the newly loaded file to flash:/startup.cfg.

        Args:
            stage (bool): whether to stage the commands or execute
                immediately

        Returns:
            True if stage=True and staging is successful
            List of etree.Element XML responses if immediate execution
        """
        if stage:
            backup = self.device.stage_config('safety_file.cfg', "save")
            rollback = self.device.stage_config(self.basename, "rollback")
            save = self.device.stage_config('startup.cfg', "save")
            return backup and rollback and save
        else:
            responses = []
            responses.append(self.device.save('safety_file.cfg'))
            responses.append(self.device.rollback(self.basename))
            responses.append(self.device.save('startup.cfg'))
            return responses



import unittest
import mock

from pyhpecw7.features.irf import IrfPort, IrfMember, InterfaceAbsentError, IRFMemberDoesntExistError
from .base_feature_test import BaseFeatureCase

class IrfTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.irf_port = IrfPort(self.device)
        self.irf_member = IrfMember(self.device)

    @mock.patch.object(IrfPort, '_build_iface_updown')
    def test_ports_build(self, mock_iface_updown):
        expected_port_build = self.read_config_xml('irf_port_build')
        expected_activate = self.read_action_xml('irf_port_build_activate')

        expected_call_list = [
            [expected_port_build, 'edit_config'],
            ['startup.cfg', 'save'],
            [expected_activate, 'action']

        ]

        down_ifaces = set(['FortyGigE1/0/2', 'FortyGigE1/0/1', 'FortyGigE1/0/3', 'FortyGigE1/0/4'])
        up_ifaces = set(['FortyGigE1/0/1', 'FortyGigE1/0/2', 'FortyGigE1/0/3', 'FortyGigE1/0/4'])

        self.irf_port.build('1', [], [], ['FortyGigE1/0/1', 'FortyGigE1/0/2'], ['FortyGigE1/0/3', 'FortyGigE1/0/4'])

        self.assert_stage_requests_multiple(expected_call_list)
        mock_iface_updown.assert_has_calls([mock.call(down_ifaces, 'down'), mock.call(up_ifaces, 'up')])

    @mock.patch.object(IrfPort, '_build_iface_updown')
    def test_ports_build_remove(self, mock_iface_updown):
        expected_port_build = self.read_config_xml('irf_port_build_remove')
        expected_activate = self.read_action_xml('irf_port_build_activate')

        expected_call_list = [
            [expected_port_build, 'edit_config'],
            ['startup.cfg', 'save'],
            [expected_activate, 'action']

        ]

        down_ifaces = set(['FortyGigE1/0/2', 'FortyGigE1/0/1', 'FortyGigE1/0/3', 'FortyGigE1/0/4'])
        up_ifaces = set()

        self.irf_port.build('1', ['FortyGigE1/0/1', 'FortyGigE1/0/2'], ['FortyGigE1/0/3', 'FortyGigE1/0/4'], [], [])

        self.assert_stage_requests_multiple(expected_call_list)
        mock_iface_updown.assert_has_calls([mock.call(down_ifaces, 'down'), mock.call(up_ifaces, 'up')])

    def test_ports_get_config(self):
        expected_get, get_reply = self.xml_get_and_reply('irf_ports')
        self.device.get.return_value = get_reply

        expected = {'1': {'irf_p2': ['FortyGigE1/0/3', 'FortyGigE1/0/4'], 'irf_p1': ['FortyGigE1/0/1', 'FortyGigE1/0/2']}}

        result = self.irf_port.get_config()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch('pyhpecw7.features.irf.Interface')
    def test_ports_iface_updown(self, mock_interface):
        mock_iface_instance = mock_interface.return_value
        mock_iface_instance.iface_exists = True

        self.irf_port._build_iface_updown(['FortyGigE1/0/1'], 'up')
        mock_iface_instance.build.assert_called_with(admin='up', stage=True)

        mock_iface_instance.iface_exists = False
        with self.assertRaises(InterfaceAbsentError):
            self.irf_port._build_iface_updown(['FortyGigE1/0/1'], 'up')

    def test_member_get_member_config(self):
        expected_get, get_reply = self.xml_get_and_reply('irf_member_config')
        self.device.get.return_value = get_reply

        expected = {'new_member_id': '1', 'priority': '4', 'descr': 'They used to call me HP1'}

        result = self.irf_member._get_member_config('1')

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_member_build_member_config(self):
        expected_call_list = [
            ['irf member 1 description From unit test', 'cli_config'],
            ['irf member 1 renumber 2', 'cli_config'],
            ['irf member 1 priority 3', 'cli_config']
        ]

        result = self.irf_member._build_member_config(member_id='1', new_member_id='2', priority='3', descr='From unit test')

        self.assert_stage_requests_multiple(expected_call_list)
        self.assertTrue(result)

    def test_member_get_domain_config(self):
        expected_get, get_reply = self.xml_get_and_reply('irf_member_domain_config')
        self.device.get.return_value = get_reply

        expected = {'domain_id': '1'}

        result = self.irf_member._get_domain_config()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_member_get_au_config(self):
        expected_get, get_reply = self.xml_get_and_reply('irf_member_au_config')
        self.device.get.return_value = get_reply

        expected = {'auto_update': 'enable'}

        result = self.irf_member._get_auto_update_config()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_member_build_au_domain_config(self):
        result = self.irf_member._build_auto_update_domain_config('enable', '2')
        expected_call = self.read_config_xml('irf_member_au_domain')

        self.assert_stage_request(expected_call, 'edit_config')
        self.assertTrue(result)

    def test_member_get_mad_exclude(self):
        self.device.cli_display.return_value = '<HP1>display current | inc "mad exclude interface"\r\n mad exclude interface FortyGigE1/0/9\r\n mad exclude interface FortyGigE1/0/10\r\n'
        result = self.irf_member._get_mad_exclude()

        self.device.cli_display.assert_called_with('display current | inc "mad exclude interface"')
        self.assertEqual(result, {'mad_exclude': ['FortyGigE1/0/9', 'FortyGigE1/0/10']})

    @mock.patch('pyhpecw7.features.irf.Interface')
    def test_member_build_mad_exclude(self, mock_interface):
        mock_iface_instance = mock_interface.return_value
        mock_iface_instance.iface_exists = True
        mock_iface_instance.interface_name = 'FortyGigE1/0/9'

        result = self.irf_member._build_mad_exclude(['FortyGigE1/0/9'])
        self.assert_stage_request('mad exclude interface FortyGigE1/0/9', 'cli_config')

        mock_iface_instance.iface_exists = False
        with self.assertRaises(InterfaceAbsentError):
            self.irf_member._build_mad_exclude(['FortyGigE1/0/9'])

    @mock.patch.object(IrfMember, '_get_member_config')
    @mock.patch.object(IrfMember, '_get_domain_config')
    @mock.patch.object(IrfMember, '_get_mad_exclude')
    @mock.patch.object(IrfMember, '_get_auto_update_config')
    def test_member_get_config(self, mock_au, mock_mad, mock_domain, mock_member):
        mock_member.return_value = {'a': 1}
        mock_domain.return_value = {'b': 2}
        mock_mad.return_value = {'c': 3}
        mock_au.return_value = {'d': 4}

        result = self.irf_member.get_config('1')
        expected = {'a': 1, 'b': 2, 'c': 3, 'd': 4}

        self.assertEqual(result, expected)

        mock_member.return_value = None
        with self.assertRaises(IRFMemberDoesntExistError):
            self.irf_member.get_config('1')

    @mock.patch.object(IrfMember, '_build_member_config')
    @mock.patch.object(IrfMember, '_build_auto_update_domain_config')
    @mock.patch.object(IrfMember, '_build_mad_exclude')
    def test_member_build(self, mock_mad, mock_au_domain, mock_member):
        params = dict(
            auto_update='disable',
            mad_exclude=['FortyGigE1/0/9', 'FortyGigE1/0/10'],
            member_id='1',
            new_member_id='1',
            domain_id='4',
            priority='2',
            descr='Configured from unit test.'
        )

        result = self.irf_member.build(**params)

        mock_member.assert_called_with(**params)
        mock_au_domain.assert_called_with('disable', '4')
        mock_mad.assert_called_with(['FortyGigE1/0/9', 'FortyGigE1/0/10'])

        self.assertTrue(result)

    def test_remove_mad_exclude(self):
        iface_list = ['FortyGigE1/0/9', 'FortyGigE1/0/10']
        self.irf_member.remove_mad_exclude(iface_list)

        self.assert_stage_requests_multiple(
            [
                ['undo mad exclude interface FortyGigE1/0/9', 'cli_config'],
                ['undo mad exclude interface FortyGigE1/0/10', 'cli_config'],
            ]
        )











if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Copyright 2015 Jason Edelman <jedelman8@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import textfsm


def get_structured_data(template, rawtxt):
    """Returns structured data given raw text using
    TextFSM templates
    """

    path = os.path.dirname(os.path.abspath(__file__)) \
        + '/textfsm_temps/' + template

    fsm = textfsm.TextFSM(open(path))

    # an object is what is being extracted
    # based on the template, it may be one objecst or multiple
    # as is the case with neighbors, interfaces, etc.
    objects = fsm.ParseText(rawtxt)

    structured_data = []
    for each in objects:
        index = 0
        temp = {}
        for template_value in each:
            temp[fsm.header[index].lower()] = str(template_value)
            index += 1
        structured_data.append(temp)

    return structured_data



"""Manage file transfer to HPCOM7 devices.
"""
from scp import SCPClient
from lxml import etree
from pyhpecw7.utils.xml.lib import *
from pyhpecw7.features.errors import FileNotEnoughSpaceError,\
    FileNotReadableError, FileRemoteDirDoesNotExist, FileTransferError, FileHashMismatchError
from pyhpecw7.errors import NCError

import paramiko
import hashlib
import re
import os


class FileCopy(object):
    """This class is used to copy local files to a ``HPCOM7`` device.

    Note:
        SCP should first be enabled on the device.

    Note:
        When using this class, the passed in ``HPCOM7`` object should
        be constructed with the ``timeout`` equal to at least 60 seconds.
        Remote MD5 sum calculations can take some time.

    Note:
        If the remote directory doesn't exist (check ``remote_dir_exists``),
        it's the responsibility of the user to call ``create_remote_dir()``
        before calling ``transfer_file()``. Otherwise, a
        ``FileRemoteDirDoesNotExist`` exception will be raised.

    Args:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.
        src (str): Full path to local file to be copied.
        dst (str): OPTIONAL - Full path or filename of remote file.
            If just a filename is supplied, 'flash:/' will be prepended.
            If nothing is supplied, the source filename will be used,
            and 'flash:/' will be prepended.
        port (int): OPTIONAL - The SSH port over which
            the SCP connection is made. Defaults to 22.

    Attributes:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.
        src (str): Full path to local file to be copied.
        dst (str): Full path of remote file.
        port (int): The SSH port over which
            the SCP connection is made.
        remote_dir_exists (bool): Whether there remote
            directory exists.
    """
    def __init__(self, device, src, dst=None, port=22):
        self.device = device
        self.src = src
        self.dst = dst or os.path.basename(src)

        if self.dst.find(':/') < 0:
            self._remote_dir = 'flash:/'
            self.dst = self._remote_dir + self.dst
        else:
            self._remote_dir = '/'.join(
                self.dst.split('/')[:-1]) + '/'

        if self._remote_dir[-2:] == ':/':
            self.remote_dir_exists = True
        else:
            self.remote_dir_exists = self._remote_dir_exists()

        self.port = port

    def _get_flash_size(self):
        """Return the available space in the remote directory.
        """
        rsp = self.device.cli_display('dir ' + self._remote_dir)
        match = re.search(r'\((\d+)(\s+KB\s+free\))', rsp)

        try:
            return int(match.group(1)) * 1000
        except:
            return 0

    def _enough_space(self):
        """Check for enough space on the remote device.

        Raises:
            FileNotEnoughSpaceError: if there isn't enough space
                on the remote device.
        """
        flash_size = self._get_flash_size()
        file_size = os.path.getsize(self.src)
        if file_size > flash_size:
            raise FileNotEnoughSpaceError(self.src, file_size, flash_size)

    def file_already_exists(self):
        """Check to see if there is a remote file with the same
        name and md5 sum.

        Returns:
            ``True`` if exists, ``False`` otherwise.
        """
        if not self.remote_dir_exists:
            return False

        try:
            dst_hash = self._get_remote_md5()
        except NCError:
            return False

        src_hash = self._get_local_md5()
        if src_hash == dst_hash:
            return True

        return False

    def _safety_checks(self):
        """Check to make sure the source file exists,
        and that there's enough space on the device.

        Throws:
            FileNotReadableError: if the local file doesn't
                exist or isn't readable.
        """
        try:
            f = open(self.src, "rb")
        except IOError:
            raise FileNotReadableError(self.src)
        finally:
            try:
                f.close()
            except NameError:
                pass

        if not self.remote_dir_exists:
            raise FileRemoteDirDoesNotExist(self._remote_dir)

        if self.remote_dir_exists:
            if not self.file_already_exists():
                self._enough_space()

    def _get_remote_md5(self):
        """Return the md5 sum of the remote file,
        if it exists.
        """
        E = action_element_maker()
        top = E.top(
            E.FileSystem(
                E.Files(
                    E.File(
                        E.SrcName(self.dst),
                        E.Operations(
                            E.md5sum()
                        )
                    )
                )
            )
        )


        nc_get_reply = self.device.action(top)
        reply_ele = etree.fromstring(nc_get_reply.xml.encode('ascii'))
        md5sum = find_in_action('md5sum', reply_ele)

        if md5sum is not None:
            return md5sum.text.strip()

    def _get_local_md5(self, blocksize=2**20):
        """Get the md5 sum of the local file,
        if it exists.
        """
        m = hashlib.md5()
        with open(self.src, "rb") as f:
            buf = f.read(blocksize)
            while buf:
                m.update(buf)
                buf = f.read(blocksize)
        return m.hexdigest()

    def _remote_dir_exists(self):
        """Check to see if the remote directory exists.
        """
        E = data_element_maker()
        top = E.top(
            E.FileSystem(
                E.Files(
                    E.File(
                        E.Name(self._remote_dir.rstrip('/')),
                        E.IsDirectory()
                    )
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))

        reply_ele = etree.fromstring(nc_get_reply.xml.encode('ascii'))
        is_dir = find_in_data('IsDirectory', reply_ele)

        if is_dir is not None\
                and is_dir.text == 'true':
            return True

        return False

    def create_remote_dir(self):
        """Create the remote directory.

        Raises:
            FileCreateDirectoryError: if the directory could
                not be created.
        """
        E = action_element_maker()
        top = E.top(
            E.FileSystem(
                E.Files(
                    E.File(
                        E.SrcName(self._remote_dir.strip('/')),
                        E.Operations(
                            E.MkDir()
                        )
                    )
                )
            )
        )

        nc_get_reply = self.device.action(top)
        reply_ele = etree.fromstring(nc_get_reply.xml.encode('ascii'))

        self.remote_dir_exists = True

    def transfer_file(self, hostname=None, username=None, password=None):
        """Transfer the file to the remote device over SCP.

        Note:
            If any arguments are omitted, the corresponding attributes
            of the ``self.device`` will be used.

        Args:
            hostname (str): OPTIONAL - The name or
                IP address of the remote device.
            username (str): OPTIONAL - The SSH username
                for the remote device.
            password (str): OPTIONAL - The SSH password
                for the remote device.

        Raises:
            FileTransferError: if an error occurs during the file transfer.
            FileHashMismatchError: if the source and
                destination hashes don't match.
            FileNotReadableError: if the local file doesn't exist or isn't readable.
            FileNotEnoughSpaceError: if there isn't enough space on the device.
            FileRemoteDirDoesNotExist: if the remote directory doesn't exist.
        """
        self._safety_checks()

        hostname = hostname or self.device.host
        username = username or self.device.username
        password = password or self.device.password

        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(
            hostname=hostname,
            username=username,
            password=password,
            port=self.port,
            allow_agent=False,
            look_for_keys=False)

        scp = SCPClient(ssh.get_transport())
        try:
            scp.put(self.src, self.dst)
        except:
            raise FileTransferError

        scp.close()

        src_hash = self._get_local_md5()
        dst_hash = self._get_remote_md5()

        if src_hash != dst_hash:
            raise FileHashMismatchError(self.src, self.dst, src_hash, dst_hash)



"""Gather LLDP neighbor information from HPCOM7 devices.
"""
from lxml import etree
from ncclient.xml_ import qualify
from pyhpecw7.utils.xml.namespaces import HPDATA, HPDATA_C
from pyhpecw7.utils.xml.lib import *


class Neighbors(object):
    """Gather LLDP neighbor information from a HP COM7 switch.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        lldp (lldp): dictionary containing LLDP neighbors
        cdp (cdp): dictionary containing CDP neighbors

    """
    def __init__(self, device):

        self.device = device

        self.lldp = self._get_neighbors(ntype='lldp')
        self.cdp = self._get_neighbors(ntype='cdp')

    def _get_interface_from_index(self, index):
        """ Returns interface name based on a given ifindex
        """
        E = data_element_maker()
        top = E.top(
            E.Ifmgr(
                E.Interfaces(
                    E.Interface(
                        E.IfIndex(index),
                        E.Name()
                    )
                )
            )
        )
        nc_get_reply = self.device.get(('subtree', top))
        interface_name = find_in_data('Name', nc_get_reply.data_ele).text

        return interface_name

    def refresh(self):
        """Refreshes the "ldp" and "cdp" attributes of the class
        """

        self.lldp = self._get_neighbors(ntype='lldp')
        self.cdp = self._get_neighbors(ntype='cdp')

    def _get_neighbors(self, ntype='lldp'):
        """Gets neighbors of device (HPCOM7)

            Args:
                ntype (str): must be "lldp" or "cdp"

            Returns:
                List of dicts with the following k/v pairs:
                    :local_intf (str): local interface of HP device
                    :neighbor_intf (str): remote interface of the neighbor
                        device
                    :neighbor (str): hostname of the neighbor device for lldp
                        and mgmt IP addr when cdp
        """
        E = data_element_maker()
        if ntype == 'cdp':
            top = E.top(
                E.LLDP(
                    E.CDPNeighbors(
                        E.CDPNeighbor()
                    )
                )
            )
        else:
            top = E.top(
                E.LLDP(
                    E.LLDPNeighbors(
                        E.LLDPNeighbor()
                    )
                )
            )

        nc_get_reply = self.device.get(('subtree', top))
        return self._build_response(nc_get_reply.data_ele, ntype=ntype)

    def _build_response(self, nc_reply, ntype='lldp'):
        """Builds dictionary from XML response coming from device

        Args:
            nc_reply (lxml.etree._Element): NETCONF
                response from device as etree element

        Returns:
            It returns a list of dictionary objects and each neighbor
            is represented as a dictionary object with the following k/v
            pairs:
                :local_intf (str): local interface of HP device
                :neighbor_intf (str): remote interface of the neighbor
                    device
                :neighbor (str): hostname of the neighbor device for lldp
                    and mgmt IP addr when cdp
        """

        if ntype == 'lldp':
            key_map = {
                'neighbor': 'SystemName',
                'neighbor_intf': 'PortId',
            }
            neighbors = findall_in_data('LLDPNeighbor', nc_reply)
        else:
            key_map = {
                'neighbor': 'ManageAdress',
                'neighbor_intf': 'PortId',
            }
            neighbors = findall_in_data('CDPNeighbor', nc_reply)

        return_neigh = []

        for neigh in neighbors:
            temp = {}
            index = find_in_data('IfIndex', neigh).text
            interface = self._get_interface_from_index(index)
            temp['local_intf'] = interface
            for new_key, xml_tag in key_map.items():
                obj = find_in_data(xml_tag, neigh)
                if obj is not None:
                    value = obj.text
                    temp[new_key] = value
            return_neigh.append(temp)

        return return_neigh



import unittest
import mock

from pyhpecw7.features.reboot import Reboot, RebootDateError, RebootTimeError

from .base_feature_test import BaseFeatureCase

class RebootTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.reboot = Reboot(self.device)

    def test_build(self):
        self.reboot.build(reboot=True)
        self.device.cli_display.assert_called_with(['reboot force'])

        self.reboot.build(stage=True, reboot=True)
        self.device.stage_config.assert_called_with(['reboot force'], 'cli_display')

    def test_build_delay(self):
        self.reboot.build(reboot=True, delay='20')
        self.device.cli_display.assert_called_with(['scheduler reboot delay 20'])

        self.reboot.build(stage=True, delay='20', reboot=True)
        self.device.stage_config.assert_called_with(['scheduler reboot delay 20'], 'cli_display')

    def test_build_at_time(self):
        self.reboot.build(reboot=True, time='10')
        self.device.cli_display.assert_called_with(['scheduler reboot at 10'])

        self.reboot.build(stage=True, time='10', reboot=True)
        self.device.stage_config.assert_called_with(['scheduler reboot at 10'], 'cli_display')

    def test_build_at_time_and_date(self):
        self.reboot.build(reboot=True, time='10', date='15')
        self.device.cli_display.assert_called_with(['scheduler reboot at 10 15'])

        self.reboot.build(stage=True, time='10', date='15', reboot=True)
        self.device.stage_config.assert_called_with(['scheduler reboot at 10 15'], 'cli_display')

    def test_param_check_time(self):
        result = self.reboot.param_check(time='03:02')
        self.assertEqual(result, None)

        with self.assertRaises(RebootTimeError):
            self.reboot.param_check(time='20')

        with self.assertRaises(RebootTimeError):
            self.reboot.param_check(time='0::03')

        with self.assertRaises(RebootTimeError):
            self.reboot.param_check(time='0:203')

    def test_param_check_date(self):
        result = self.reboot.param_check(date='03/02/2939')
        self.assertEqual(result, None)

        with self.assertRaises(RebootDateError):
            self.reboot.param_check(date='03.02.2939')

        with self.assertRaises(RebootDateError):
            self.reboot.param_check(date='03/02/2939/3929')

        with self.assertRaises(RebootDateError):
            self.reboot.param_check(date='003/02/2939')

        with self.assertRaises(RebootDateError):
            self.reboot.param_check(date='03/002/2939')

        with self.assertRaises(RebootDateError):
            self.reboot.param_check(date='03/02/29393')

if __name__ == "__main__":
    unittest.main()


"""Install an operating system on HPCOM7 devices.
"""
from lxml import etree
from ncclient.xml_ import qualify
from pyhpecw7.utils.xml.namespaces import HPDATA, HPDATA_C, HPACTION
from pyhpecw7.utils.xml.lib import *


class InstallOs(object):
    """This class is used to get and build the startup software image.
    It is often used in conjunction with ``file_copy.FileCopy``, which
    can transfer an image.

    Args:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.

    Attributes:
        device (HPCOM7): connected instance of
            a ``pyhpecw7.comware.HPCOM7`` object.
    """
    def __init__(self, device):
        self.device = device

    def get_config(self):
        """Return a dictionary of current and startup image names.

        Returns:
            A dictionary of 'current', 'startup-primary', and
            'startup-backup' image file names. For each, there
            is a 'boot' image and a 'system' image::
                {
                    'current': {
                        'boot' : <current boot image>,
                        'system': <current system image>,
                    }
                    'startup-primary': {
                        'boot' : <primary startup boot image>,
                        'system': <primary startup system image>,
                    }
                    'startup-backup': {
                        'boot': <backup startup boot image>,
                        'system': <backup startup system image>
                    }
                }
        """
        E = data_element_maker()
        top = E.top(
            E.Package(
                E.BootLoaderList(
                    E.BootList()
                )
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        boot_lists = findall_in_data('BootList', nc_get_reply.data_ele)

        image_dict = {}

        key_map = {'0': 'current',
                   '1': 'startup-primary',
                   '2': 'startup-backup'}

        for boot_list in boot_lists:
            list_num = boot_list.findtext(
                './/{0}{1}'.format(HPDATA_C, 'BootType'))
            image_iter = boot_list.iterfind(
                './/{0}{1}'.format(HPDATA_C, 'FileName'))
            try:
                boot_file = next(image_iter).text
                sys_file = next(image_iter).text
            except StopIteration:
                continue

            list_type = key_map[list_num]
            image_dict[list_type] = {}
            image_dict[list_type]['boot'] = boot_file.split(':/')[1]
            image_dict[list_type]['system'] = sys_file.split(':/')[1]

        return image_dict

    def build(self, os_type, ipe=None, boot=None, system=None,
              delete_ipe=False, stage=False):
        """Stage or execute the configuration commands
        for changing the primary startup image.

        Args:
            os_type (str): REQUIRED - 'ipe' (for IPE packages)
                or 'bootsys' (for separate boot and system files.
            ipe (str): REQUIRED if ``os_type`` is 'ipe'
                - The full path of the remote IPE file.
            boot (str): REQUIRED if ``os_type`` is 'bootsys'
                - The full path of the remote boot .bin file.
            system (str): REQUIRED if ``os_type`` is 'bootsys'
                - The full path of the remote system .bin file.
            delete_ipe (bool): OPTIONAL - Whether to delete the remote IPE file
                after config change. Defaults to ``False``.
            stage (bool): whether to stage the commands or execute
                immediately

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediate execution
        """
        E = action_element_maker()

        ipe_params = []
        boot_sys_params = []
        if os_type == 'ipe':
            ipe_params.append(E.IPEFileName(ipe))
            ipe_params.append(E.DeleteIPEFile(str(delete_ipe).lower()))
        elif os_type == 'bootsys':
            boot_sys_ele = E.ImageFiles(
                E.Boot(boot),
                E.System(system)
            )
            boot_sys_params.append(boot_sys_ele)

        top = E.top(
            E.Package(
                E.SetBootImage(
                    E.Type('1'),
                    E.OverwriteLocalFile('false'),
                    *ipe_params
                ),
                *boot_sys_params
            )
        )

        if stage:
            return self.device.stage_config(top, 'action')
        else:
            return self.device.action(top)



import unittest
import mock
from lxml import etree

from pyhpecw7.features.ipinterface import V4, V6, IpInterface
from pyhpecw7.features.errors import IpIfaceMissingData

from .base_feature_test import BaseFeatureCase

IFACE_INDEX = '9'

class IpInterfaceTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    @mock.patch('pyhpecw7.features.ipinterface.Interface')
    def setUp(self, mock_iface, mock_device):
        self.device = mock_device

        mock_iface.return_value.iface_index = IFACE_INDEX
        mock_iface.return_value.is_routed = True

        self.ip4interface = IpInterface(self.device, 'FortyGigE1/0/3')
        self.ip6interface = IpInterface(self.device, 'FortyGigE1/0/3', version=V6)


    def test_init(self):
        self.assertEqual(self.ip4interface.is_routed, True)
        self.assertEqual(self.ip4interface.interface.iface_index, IFACE_INDEX)
        self.assertEqual(self.ip4interface.version, V4)

        self.assertEqual(self.ip6interface.is_routed, True)
        self.assertEqual(self.ip6interface.interface.iface_index, IFACE_INDEX)
        self.assertEqual(self.ip6interface.version, V6)


    def test_get_config_ipv4(self):
        expected_get, get_reply = self.xml_get_and_reply('ipinterface_v4')
        self.device.get.return_value = get_reply

        expected = [{'mask': '255.255.255.0', 'addr': '192.168.3.5'}]

        ipint = self.ip4interface.get_config()

        self.assertEqual(ipint, expected)
        self.assert_get_request(expected_get)

    def test_get_config_ipv6(self):
        expected_get, get_reply = self.xml_get_and_reply('ipinterface_v6')
        self.device.get.return_value = get_reply

        expected = [{'mask': '10', 'addr': '2001:DB8::1'}, {'mask': '10', 'addr': '2001:DB8::2'}, {'mask': '10', 'addr': 'FE80::4631:92FF:FE7C:C866'}]

        ipint = self.ip6interface.get_config()

        self.assertEqual(ipint, expected)
        self.assert_get_request(expected_get)

    def test_build_config(self):
        # V4 Tests

        result = self.ip4interface._build_config('present', addr='1.2.3.4', mask='255.255.255.0')
        expected = self.read_config_xml('ipint_v4_basic')
        self.assert_elements_equal(result, expected)

        result = self.ip4interface._build_config('present', addr='1.2.3.4', mask='24')
        expected = self.read_config_xml('ipint_v4_basic')
        self.assert_elements_equal(result, expected)

        with self.assertRaises(IpIfaceMissingData):
            self.ip4interface._build_config('present', addr='1.2.3.4')

        result = self.ip4interface._build_config('absent', addr='1.2.3.4', mask='24')
        expected = self.read_config_xml('ipint_v4_absent')
        self.assert_elements_equal(result, expected)

        result = self.ip4interface._build_config('absent', addr='1.2.3.4', mask='255.255.255.0')
        expected = self.read_config_xml('ipint_v4_absent')
        self.assert_elements_equal(result, expected)


        # V6 Tests

        result = self.ip6interface._build_config('present', addr='2001:DB8::1', mask='10')
        expected = self.read_config_xml('ipint_v6_basic')
        self.assert_elements_equal(result, expected)

        result = self.ip6interface._build_config('absent', addr='2001:DB8::1', mask='10')
        expected = self.read_config_xml('ipint_v6_absent')
        self.assert_elements_equal(result, expected)


    @mock.patch.object(IpInterface, '_build_config')
    def test_build(self, mock_build_config):
        self.ip4interface.build(addr='1.2.3.4', mask='24')
        mock_build_config.assert_called_with(state='present', addr='1.2.3.4', mask='24')
        self.device.edit_config.assert_called_with(mock_build_config.return_value)

        self.ip4interface.build(stage=False, addr='1.2.3.4', mask='24')
        mock_build_config.assert_called_with(state='present', addr='1.2.3.4', mask='24')
        self.device.edit_config.assert_called_with(mock_build_config.return_value)

        self.ip4interface.build(stage=True, addr='1.2.3.4', mask='24')
        mock_build_config.assert_called_with(state='present', addr='1.2.3.4', mask='24')
        self.device.stage_config.assert_called_with(mock_build_config.return_value, 'edit_config')

    @mock.patch.object(IpInterface, '_build_config')
    def test_remove(self, mock_build_config):
        self.ip4interface.remove(addr='1.2.3.4', mask='24')
        mock_build_config.assert_called_with(state='absent', addr='1.2.3.4', mask='24')
        self.device.edit_config.assert_called_with(mock_build_config.return_value)

        self.ip4interface.remove(stage=False, addr='1.2.3.4', mask='24')
        mock_build_config.assert_called_with(state='absent', addr='1.2.3.4', mask='24')
        self.device.edit_config.assert_called_with(mock_build_config.return_value)

        self.ip4interface.remove(stage=True, addr='1.2.3.4', mask='24')
        mock_build_config.assert_called_with(state='absent', addr='1.2.3.4', mask='24')
        self.device.stage_config.assert_called_with(mock_build_config.return_value, 'edit_config')


if __name__ == '__main__':
    unittest.main()


import unittest
import mock

from pyhpecw7.features.vxlan import Vxlan, Tunnel, L2EthService
from .base_feature_test import BaseFeatureCase

INTERFACE = 'FortyGigE1/0/2'
INSTANCE = '100'
VSI = 'VSI_VXLAN_100'
TUNNEL = '20'

class VxlanTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self,  mock_device):
        self.device = mock_device
        self.l2eth = L2EthService(self.device, INTERFACE, INSTANCE, VSI)
        self.vxlan = Vxlan(self.device, INSTANCE, vsi=VSI)
        self.tunnel = Tunnel(self.device, TUNNEL)


    def test_tunnel_get_config(self):
        self.device.cli_display.return_value = """<HP1>display current-configuration interface Tunnel 20
#
interface Tunnel20 mode vxlan
 source 10.1.1.1
 destination 10.1.1.2
#
return
"""
        expected = {'dest': '10.1.1.2', 'src': '10.1.1.1', 'mode': 'vxlan'}
        result = self.tunnel.get_config()

        self.assertEqual(result, expected)

    def test_tunnel_get_config_no_tunnel(self):
        self.device.cli_display.return_value = """                                                    ^
 % Wrong parameter found at '^' position.
"""
        expected = {}
        result = self.tunnel.get_config()

        self.assertEqual(result, expected)

    def test_tunnel_get_global_source(self):
        self.device.cli_display.return_value = """<HP1>display current-configuration | inc "tunnel global source"
 tunnel global source-address 10.10.10.10

"""
        expected = '10.10.10.10'
        result = self.tunnel.get_global_source()

        self.assertEqual(result, expected)

    def test_tunnel_build_config(self):
        self.tunnel._build_config('present', src='1.1.1.1', dest='2.2.2.2', global_src='1.1.1.2')
        expected_call = ['tunnel global source-address 1.1.1.2', 'interface tunnel 20 mode vxlan', 'source 1.1.1.1', 'destination 2.2.2.2']

        self.device.cli_config.assert_called_with(expected_call)

    def test_tunnel_build_config_stage(self):
        self.tunnel._build_config('present', src='1.1.1.1', dest='2.2.2.2', global_src='1.1.1.2', stage=True)
        expected_call = ['tunnel global source-address 1.1.1.2', 'interface tunnel 20 mode vxlan', 'source 1.1.1.1', 'destination 2.2.2.2']

        self.device.stage_config.assert_called_with(expected_call, 'cli_config')

    def test_tunnel_build_config_remove(self):
        self.tunnel._build_config('absent', src='1.1.1.1', dest='2.2.2.2', global_src='1.1.1.2')
        expected_call = ['undo interface tunnel 20']

        self.device.cli_config.assert_called_with(expected_call)

    @mock.patch.object(Tunnel, '_build_config')
    def test_tunnel_remove_stage(self, mock_build_config):
        self.tunnel.remove(stage=True)
        mock_build_config.assert_called_with(state='absent', stage=True)

    @mock.patch.object(Tunnel, '_build_config')
    def test_tunnel_remove(self, mock_build_config):
        self.tunnel.remove()
        mock_build_config.assert_called_with(state='absent', stage=False)

    @mock.patch.object(Tunnel, '_build_config')
    def test_tunnel_build_stage(self, mock_build_config):
        self.tunnel.build(stage=True)
        mock_build_config.assert_called_with(state='present', stage=True)

    @mock.patch.object(Tunnel, '_build_config')
    def test_tunnel_build(self, mock_build_config):
        self.tunnel.build()
        mock_build_config.assert_called_with(state='present', stage=False)

    @mock.patch.object(Vxlan, 'get_tunnels')
    def test_get_config(self, mock_get_tunnels):
        mock_get_tunnels.return_value = ['20']

        expected_get, get_reply = self.xml_get_and_reply('vxlan')
        self.device.get.return_value = get_reply

        expected = {'vsi': 'VSI_VXLAN_100', 'tunnels': ['20'], 'vxlan': '100'}

        result = self.vxlan.get_config()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_build_vsi(self):
        result = self.vxlan._build_vsi('merge')
        expected = self.read_config_xml('vxlan_vsi')

        self.assert_elements_equal(result, expected)

    def test_build_vxlan(self):
        result = self.vxlan._build_vxlan('merge')
        expected = self.read_config_xml('vxlan')

        self.assert_elements_equal(result, expected)

    def test_build_tunnels(self):
        result = self.vxlan._build_tunnels('merge', ['20', '3'])
        expected = self.read_config_xml('vxlan_tunnels')

        self.assert_elements_equal(result, expected)

    def test_build(self):
        result = self.vxlan.build(tunnels_to_add=['20', '3'])
        expected_add_call = self.read_config_xml('vxlan_tunnels')

        self.assert_config_request(expected_add_call)

        result = self.vxlan.build(tunnels_to_remove=['4', '5'], stage=True)
        expected_rmv_call = self.read_config_xml('vxlan_tunnels_delete')

        self.assert_stage_request(expected_rmv_call, 'edit_config')
        self.assertEqual(result, True)

    def test_create(self):
        expected_vsi_call = self.read_config_xml('vxlan_vsi')
        expected_vxlan_call = self.read_config_xml('vxlan')

        result = self.vxlan.create()
        self.assert_config_request(expected_vxlan_call)

        result = self.vxlan.create(stage=True)
        self.assert_stage_request(expected_vxlan_call, 'edit_config')

    def test_remove_vsi(self):
        expected_vsi_call = self.read_config_xml('vxlan_vsi_delete')
        result = self.vxlan.remove_vsi()

        self.assert_config_request(expected_vsi_call)
        self.assertTrue(result)

        result = self.vxlan.remove_vsi(stage=True)
        self.assert_stage_request(expected_vsi_call, 'edit_config')

    def test_remove_vxlan(self):
        expected_call = self.read_config_xml('vxlan_delete')
        result = self.vxlan.remove_vxlan()

        self.assert_config_request(expected_call)
        self.assertTrue(result)

        result = self.vxlan.remove_vxlan(stage=True)
        self.assert_stage_request(expected_call, 'edit_config')

    def test_get_tunnels(self):
        expected_get, get_reply = self.xml_get_and_reply('vxlan_tunnels')
        self.device.get.return_value = get_reply

        expected = ['20']

        result = self.vxlan.get_tunnels()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_l2eth_vsi_exist(self):
        expected_get, get_reply = self.xml_get_and_reply('l2eth_vsi_exist')
        self.device.get.return_value = get_reply

        expected = {'vsi': 'VSI_VXLAN_100'}

        result = self.l2eth.vsi_exist()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(L2EthService, '_index_from_interface')
    def test_l2eth_get_vsi_map(self, mock_index_from_interface):
        mock_index_from_interface.return_value = '125'

        expected_get, get_reply = self.xml_get_and_reply('l2eth_vsi_map')
        self.device.get.return_value = get_reply

        expected = {'vsi': 'VSI_VXLAN_100', 'access_mode': 'ethernet', 'index': '125', 'instance': '100'}

        result = self.l2eth.get_vsi_map()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(L2EthService, '_index_from_interface')
    @mock.patch.object(L2EthService, '_get_interface_from_index')
    def test_l2eth_get_vsi_encap(self, mock_iface_from_index, mock_index_from_interface):
        mock_iface_from_index.return_value = 'FortyGigE1/0/32'
        mock_index_from_interface.return_value = '125'

        expected_get, get_reply = self.xml_get_and_reply('l2eth_vsi_encap')
        self.device.get.return_value = get_reply

        expected = {'interface': 'FortyGigE1/0/32', 'index': '125', 'encap': 'tagged', 'instance': '100'}

        result = self.l2eth.get_vsi_encap()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)


    @mock.patch.object(L2EthService, '_index_from_interface')
    def test_l2eth_build_encap_default(self, mock_index_from_interface):
        mock_index_from_interface.return_value = '125'
        expected_call = self.read_config_xml('l2eth_build_encap_default')

        self.l2eth._build_encap('merge', encap='default')
        self.assert_config_request(expected_call)

        self.l2eth._build_encap('merge', stage=True, encap='default')
        self.assert_stage_request(expected_call, 'edit_config')

    @mock.patch.object(L2EthService, '_index_from_interface')
    def test_l2eth_build_encap_tagged(self, mock_index_from_interface):
        mock_index_from_interface.return_value = '125'
        expected_call = self.read_config_xml('l2eth_build_encap_tagged')

        self.l2eth._build_encap('merge', encap='tagged')
        self.assert_config_request(expected_call)

        self.l2eth._build_encap('merge', stage=True, encap='tagged')
        self.assert_stage_request(expected_call, 'edit_config')

    @mock.patch.object(L2EthService, '_index_from_interface')
    def test_l2eth_build_encap_only_tagged(self, mock_index_from_interface):
        mock_index_from_interface.return_value = '125'
        expected_call = self.read_config_xml('l2eth_build_encap_only_tagged')

        self.l2eth._build_encap('merge', encap='only-tagged', vlanid='10')
        self.assert_config_request(expected_call)

        self.l2eth._build_encap('merge', stage=True, encap='only-tagged', vlanid='10')
        self.assert_stage_request(expected_call, 'edit_config')

    @mock.patch.object(L2EthService, '_index_from_interface')
    def test_l2eth_build_encap_only_delete(self, mock_index_from_interface):
        mock_index_from_interface.return_value = '125'
        expected_call = self.read_config_xml('l2eth_build_encap_delete')

        self.l2eth._build_encap('delete')
        self.assert_config_request(expected_call)

        self.l2eth._build_encap('delete', stage=True)
        self.assert_stage_request(expected_call, 'edit_config')

    def test_l2eth_build_xconnect(self):
        expected_call = self.read_config_xml('l2eth_build_xconnect')

        self.l2eth._build_xconnect('merge', '125', access_mode='ethernet')
        self.assert_config_request(expected_call)

        self.l2eth._build_xconnect('merge', '125', stage=True, access_mode='ethernet')
        self.assert_stage_request(expected_call, 'edit_config')

    @mock.patch.object(L2EthService, '_build_encap')
    @mock.patch.object(L2EthService, '_build_xconnect')
    def test_l2eth_build_config(self, mock_xconnect, mock_encap):
        self.l2eth.jindex = '999'

        self.l2eth._build_config('present', encap='tagged', access_mode='vlan')
        mock_encap.assert_called_with('merge', access_mode='vlan', encap='tagged', stage=False)
        mock_xconnect.assert_called_with('merge', '999', access_mode='vlan', encap='tagged', stage=False)

    @mock.patch.object(L2EthService, '_build_encap')
    @mock.patch.object(L2EthService, '_build_xconnect')
    def test_l2eth_build_config_absent(self, mock_xconnect, mock_encap):
        self.l2eth.jindex = '999'

        self.l2eth._build_config('absent', encap='tagged', access_mode='vlan')
        mock_encap.assert_called_with('delete', access_mode='vlan', encap='tagged', stage=False)
        mock_xconnect.assert_not_called()

    @mock.patch.object(L2EthService, '_build_config')
    def test_l2eth_build(self, mock_build_config):
        self.l2eth.build(vsi='abc', encap='tagged', access_mode='ethernet')
        mock_build_config.assert_called_with(access_mode='ethernet', encap='tagged', stage=False, state='present', vsi='abc')

    @mock.patch.object(L2EthService, '_build_config')
    def test_l2eth_remove(self, mock_build_config):
        self.l2eth.remove()
        mock_build_config.assert_called_with(stage=False, state='absent')

    def test_l2eth_get_interface_from_index(self):
        expected_get, get_reply = self.xml_get_and_reply('l2eth_get_interface_from_index')
        self.device.get.return_value = get_reply

        expected = 'FortyGigE1/0/32'

        result = self.l2eth._get_interface_from_index('125')

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)


if __name__ == '__main__':
    unittest.main()



"""Manage L2VPN on HPCOM7 devices.
"""

from pyhpecw7.utils.xml.lib import *


class L2VPN(object):
    """Enable/Disable L2VPN globally on a HP Comware 7 switch.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.

    """
    def __init__(self, device):
        self.device = device

    def get_config(self):
        """Get current L2VPN global configuration state.
        """
        KEYMAP = {
            'enable': 'Enable',
            'vsi_supported': 'SupportVsiInterface'
        }

        VALUE_MAP = {
            'Enable': {
                'true': 'enabled',
                'false': 'disabled'
            }
        }

        E = data_element_maker()
        top = E.top(
            E.L2VPN(
                E.Base()
            )
        )

        nc_get_reply = self.device.get(('subtree', top))
        return_l2vpn = data_elem_to_dict(nc_get_reply.data_ele, KEYMAP, value_map=VALUE_MAP)

        return return_l2vpn.get('enable')

    def enable(self, stage=False):
        """Stage or execute a config object to enable L2VPN

        Args:
            stage (bool): whether to stage the commands or execute
                immediately

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediate execution
        """
        config = self._build_config(state='enabled')
        if stage:
            return self.device.stage_config(config, 'edit_config')
        else:
            return self.device.edit_config(config)

    def disable(self, stage=False):
        """Stage or execute a config object to disable L2VPN

        Args:
            stage (bool): whether to stage the commands or execute
                immediately

        Returns:
            True if stage=True and successfully staged
            etree.Element XML response if immediate execution
        """
        config = self._build_config(state='disabled')
        if stage:
            return self.device.stage_config(config, 'edit_config')
        else:
            return self.device.edit_config(config)

    def _build_config(self, state):
        """Build config object to configure L2VPN global features

        Args:
            state (str): must be "enabled" or "disabled" and is the desired
                state of the L2VPN global feature

        Returns:
            etree.Element config object to configure L2VPN global features
        """
        if state == 'enabled':
            value = 'true'
        elif state == 'disabled':
            value = 'false'

        EN = nc_element_maker()
        EC = config_element_maker()

        config = EN.config(
            EC.top(
                EC.L2VPN(
                    EC.Base(
                        EC.Enable(value),
                        **operation_kwarg('merge')
                    )
                )
            )
        )

        return config



from distutils.core import setup
from setuptools import find_packages

setup(
  name='py3hpecw7',
  packages=find_packages(),
  version='0.1.0',
  description='Python package to simplify working with HPE Comware7 based devices',
  author='HPE',
  license='Apache2',
  url='https://github.com/vincentbernat/pyhpecw7',
  classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
  keywords='HPE Comware7 FlexFabric Netconf API ',

  package_data={'pyhpecw7': ['utils/templates/textfsm_temps/*.tmpl']},
  install_requires=[
      'textfsm>=0.2.1',
      'lxml',
      'ncclient',
      'scp',
      'ipaddr>=2.1.11',
      'paramiko'
  ],
)



"""Gather device facts on HPCOM7 devices.
"""
import collections
from pyhpecw7.utils.xml.lib import *


class Facts(object):
    """Gather device facts from a HP Comware 7 device.

    Args:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.

    Attributes:
        device (HPCOM7): connected instance of a ``pyhpecw7.comware.HPCOM7``
            object.
        facts (dict): this is a read-only attribute. Details can be seen
            in ``get_facts``.

    """
    def __init__(self, device):
        self.device = device
        self.em = data_element_maker()

    @property
    def facts(self):
        return self.get_facts()

    def get_facts(self):
        """Gather facts from the HP Comware 7 device

        Returns:
            This returns a dictionary with several key/value
            pairs describing the device.  See example below.

            Example::

                {
                    'hostname': 'HP5930_1',
                    'interface_list': ["FortyGigE1/0/1", "FortyGigE1/0/2",
                        "FortyGigE1/0/3", "FortyGigE1/0/4"],
                    'localtime': '2016-01-30T01:47:07',
                    'model': 'HP FF 5930-32QSFP+ Switch',
                    'os': '7.1.045 Release 2418P01',
                    'serial_number': 'CN43G9800T',
                    'uptime': '6d 23hr 25min 18sec',
                    'vendor': 'hp',
                    'hardware': 'Ver.A'
                }
        """

        facts = collections.OrderedDict()
        facts.update(self._get_inventory())
        facts.update(self._get_base())
        facts.update(self._get_interface_list())

        return facts

    def _get_interface_list(self):
        """Get interface list that will be added to facts.
        """
        E = self.em
        top = E.top(
                E.Ifmgr(
                    E.Interfaces(
                        E.Interface(
                            E.Name()
                        )
                    )
                )
            )

        nc_get_reply = self.device.get(('subtree', top))

        intfs_xml = findall_in_data('Name', nc_get_reply.data_ele)
        interfaces = [intf.text for intf in intfs_xml]
        intfs = dict(interface_list=interfaces)

        return intfs

    def _get_uptime(self, seconds):
        """Convert seconds to d, hr, min, sec format.
        """
        m, s = divmod(int(seconds), 60)
        h, m = divmod(m, 60)
        d = 0
        if h > 23:
            d, h = divmod(h, 24)
        uptime = '{0}d {1}hr {2}min {3}sec'.format(d, h, m, s)

        return uptime

    def _get_inventory(self):
        """Get os, serial number, and model that will be added to facts.
        """
        key_map = {
            'os': 'SoftwareRev',
            'serial_number': 'SerialNum',
            'model': 'ModelName',
            'hardware': 'HardwareRev'
        }

        E = self.em
        top = E.top(
            E.LLDP(
                E.Inventory(
                    E.SoftwareRev(),
                    E.SerialNum(),
                    E.ModelName(),
                    E.HardwareRev(),

                )
            )
        )
        nc_get_reply = self.device.get(('subtree', top))

        inventory = data_elem_to_dict(nc_get_reply.data_ele, key_map)
        inventory['vendor'] = 'hp'

        return inventory

    def _get_base(self):
        """Get hostname, localtime, and uptime that will be added to facts.
        """
        key_map = {
            'hostname': 'HostName',
            'localtime': 'LocalTime',
            'uptime': 'Uptime'
        }

        E = self.em
        top = E.top(
            E.Device(
                E.Base(
                    E.HostName(),
                    E.LocalTime(),
                    E.Uptime(),
                )
            )
        )
        nc_get_reply = self.device.get(('subtree', top))

        basefacts = data_elem_to_dict(nc_get_reply.data_ele, key_map)
        basefacts['uptime'] = self._get_uptime(basefacts.get('uptime', 0))

        return basefacts



import unittest
import mock

from pyhpecw7.features.switchport import Switchport

from .base_feature_test import BaseFeatureCase

IFACE_INDEX = '9'
IFACE_NAME = 'FortyGigE1/0/3'


class SwitchportTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    @mock.patch('pyhpecw7.features.switchport.Interface')
    def setUp(self, mock_iface, mock_device):
        self.device = mock_device

        mock_iface.return_value.iface_index = IFACE_INDEX
        mock_iface.return_value.is_routed = True
        mock_iface.return_value.interface_name = IFACE_NAME

        self.switchport = Switchport(self.device, IFACE_NAME)

    def test_init(self):
        self.assertEqual(self.switchport.interface_name, IFACE_NAME)
        self.assertEqual(self.switchport.link_type, 'unknown')

    def test_get_config(self):
        expected_get, get_reply = self.xml_get_and_reply('switchport')
        self.device.get.return_value = get_reply

        expected = {'pvid': '1', 'link_type': 'access'}

        result = self.switchport.get_config()

        self.assertEqual(self.switchport.link_type, 'access')
        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_get_config_none(self):
        expected_get, get_reply = self.xml_get_and_reply('switchport_none')
        self.device.get.return_value = get_reply

        expected = {}

        result = self.switchport.get_config()

        self.assertEqual(self.switchport.link_type, 'unknown')
        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    @mock.patch.object(Switchport, 'build')
    def test_default(self, mock_build):
        expected_args = {'pvid': '1', 'link_type': 'access'}
        self.switchport.default()

        mock_build.assert_called_with(stage=False, **expected_args)

    def test_convert_interface_access(self):
        expected = self.read_config_xml('switchport_convert_access')
        self.switchport.convert_interface('access')
        self.assert_config_request(expected)

        self.switchport.convert_interface('access', stage=True)
        self.assert_stage_request(expected, 'edit_config')

    def test_convert_interface_trunk(self):
        expected = self.read_config_xml('switchport_convert_trunk')
        self.switchport.convert_interface('trunk')
        self.assert_config_request(expected)

        self.switchport.convert_interface('trunk', stage=True)
        self.assert_stage_request(expected, 'edit_config')

    def test_build_access(self):
        expected = self.read_config_xml('switchport_build_access')
        self.switchport.build(link_type='access', pvid='2')
        self.assert_config_request(expected)

        self.switchport.build(stage=True, link_type='access', pvid='2')
        self.assert_stage_request(expected, 'edit_config')

    def test_build_trunk(self):
        expected = self.read_config_xml('switchport_build_trunk')
        self.switchport.build(link_type='trunk', pvid='2')
        self.assert_config_request(expected)

        self.switchport.build(stage=True, link_type='trunk', pvid='2')
        self.assert_stage_request(expected, 'edit_config')


if __name__ == '__main__':
    unittest.main()



import unittest
import mock

from pyhpecw7.features.ping import Ping
from pyhpecw7.features.errors import InvalidIPAddress
from .base_feature_test import BaseFeatureCase

TARGET = '8.8.8.8'

class PingTestCase(BaseFeatureCase):

    maxDiff = None

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device

    @mock.patch.object(Ping, '_ping')
    def test_init(self, mock_ping):
        ping = Ping(self.device, TARGET)

        self.assertEqual(ping.vrf, '')
        self.assertEqual(ping.host, TARGET)
        self.assertEqual(ping.v6, False)
        self.assertEqual(ping.detail, False)

        ping6 = Ping(self.device, TARGET, v6=True)

        self.assertEqual(ping6.vrf, '')
        self.assertEqual(ping6.host, TARGET)
        self.assertEqual(ping6.v6, True)
        self.assertEqual(ping6.detail, False)

        mock_ping.assert_has_calls([mock.call(), mock.call()])

    @mock.patch.object(Ping, '_build_response')
    def test_ping(self, mock_build):
        ping = Ping(self.device, TARGET)

        action, reply = self.xml_action_and_reply('ping')
        self.device.action.return_value = reply

        result = ping._ping()

        self.assert_action_request(action)
        mock_build.assert_called_with(reply)
        self.assertEqual(result, mock_build.return_value)

    @mock.patch.object(Ping, '_ping')
    def test_build_response(self, mock_ping):
        ping = Ping(self.device, TARGET, detail=True)
        reply = self.read_action_reply_xml('ping')
        result = ping._build_response(reply)

        expected = {
            'payload_length': '56',
            'max': '8',
            'packets_rx': '5',
            'detailed_response': [
                {
                    'reply_time': '8',
                    'icmp_seq': '0'
                },
                {
                    'reply_time': '7',
                    'icmp_seq': '1'
                },
                {
                    'reply_time': '7',
                    'icmp_seq': '2'
                },
                {
                    'reply_time': '7',
                    'icmp_seq': '3'
                },
                {
                    'reply_time': '11',
                    'icmp_seq': '4'
                }
            ],
            'loss_rate': '0',
            'host': '8.8.8.8',
            'min': '8',
            'packets_tx': '5',
            'avg': '8'
        }

        self.assertEqual(result, expected)

    @mock.patch.object(Ping, '_ping')
    def test_param_check(self, mock_ping):
        ping = Ping(self.device, '1.2.3')
        with self.assertRaises(InvalidIPAddress):
            ping.param_check(host=ping.host)


if __name__ == '__main__':
    unittest.main()



import unittest
import mock
import collections

from ncclient.operations.retrieve import GetReply
from pyhpecw7.features.facts import Facts

from .base_feature_test import BaseFeatureCase

class FactsTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.facts = Facts(self.device)

    def test_get_interface_list(self):
        expected_get, get_reply = self.xml_get_and_reply('interface_list')
        self.device.get.return_value = get_reply

        iface_list = self.facts._get_interface_list()

        self.assertIn('interface_list', iface_list)
        self.assertIn('FortyGigE1/0/3', iface_list['interface_list'])

        self.assert_get_request(expected_get)

    def test_get_inventory(self):
        expected_get, get_reply = self.xml_get_and_reply('inventory')
        self.device.get.return_value = get_reply

        expected = {
            'vendor': 'hp',
            'serial_number': 'CN43G9800T',
            'model': 'HP FF 5930-32QSFP+ Switch',
            'os': '7.1.045 ESS 2415',
            'hardware': 'Ver.A'
        }

        inventory = self.facts._get_inventory()

        self.assertEqual(inventory, expected)
        self.assert_get_request(expected_get)

    def test_get_base(self):
        expected_get, get_reply = self.xml_get_and_reply('facts_base')
        self.device.get.return_value = get_reply

        expected = {
            'uptime': '2d 3hr 51min 45sec',
            'hostname': 'HP1',
            'localtime': '2011-01-03T03:50:39'
        }

        facts_base = self.facts._get_base()

        self.assertEqual(facts_base, expected)
        self.assert_get_request(expected_get)

    def test_get_uptime(self):
        a = self.facts._get_uptime(0)
        self.assertEqual(a, '0d 0hr 0min 0sec')

        b = self.facts._get_uptime(1)
        self.assertEqual(b, '0d 0hr 0min 1sec')

        c = self.facts._get_uptime(60)
        self.assertEqual(c, '0d 0hr 1min 0sec')

        d = self.facts._get_uptime(61)
        self.assertEqual(d, '0d 0hr 1min 1sec')

        e = self.facts._get_uptime(3928)
        self.assertEqual(e, '0d 1hr 5min 28sec')

        f = self.facts._get_uptime(355928)
        self.assertEqual(f, '4d 2hr 52min 8sec')


    @mock.patch.object(Facts, '_get_interface_list')
    @mock.patch.object(Facts, '_get_inventory')
    @mock.patch.object(Facts, '_get_base')
    def test_facts(self, mock_base, mock_inventory, mock_iface_list):
        mock_base.return_value = {'uptime': '2d 4hr 22min 28sec', 'hostname': 'HP1', 'localtime': '2011-01-03T04:21:22'}
        mock_inventory.return_value = {'vendor': 'hp', 'serial_number': 'CN43G9800T', 'model': 'HP FF 5930-32QSFP+ Switch', 'os': '7.1.045 ESS 2415', 'hardware': 'Ver.A'}
        mock_iface_list.return_value = {'interface_list': ['FortyGigE1/0/1', 'FortyGigE1/0/2', 'FortyGigE1/0/3', 'FortyGigE1/0/4', 'FortyGigE1/0/5', 'FortyGigE1/0/6', 'FortyGigE1/0/7', 'FortyGigE1/0/8', 'FortyGigE1/0/9', 'FortyGigE1/0/10', 'FortyGigE1/0/11', 'FortyGigE1/0/12', 'FortyGigE1/0/13', 'FortyGigE1/0/14', 'FortyGigE1/0/15', 'FortyGigE1/0/16', 'FortyGigE1/0/17', 'FortyGigE1/0/18', 'FortyGigE1/0/19', 'FortyGigE1/0/20', 'FortyGigE1/0/21', 'FortyGigE1/0/22', 'FortyGigE1/0/23', 'FortyGigE1/0/24', 'FortyGigE1/0/25', 'FortyGigE1/0/26', 'FortyGigE1/0/27', 'FortyGigE1/0/28', 'FortyGigE1/0/29', 'FortyGigE1/0/30', 'FortyGigE1/0/31', 'FortyGigE1/0/32', 'M-GigabitEthernet0/0/0', 'NULL0', 'InLoopBack0', 'Register-Tunnel0', 'LoopBack29', 'Bridge-Aggregation100']}

        expected = collections.OrderedDict(
            [
                ('vendor', 'hp'),
                ('serial_number', 'CN43G9800T'),
                ('model', 'HP FF 5930-32QSFP+ Switch'),
                ('os', '7.1.045 ESS 2415'),
                ('hardware', 'Ver.A'),
                ('uptime', '2d 4hr 22min 28sec'),
                ('hostname', 'HP1'),
                ('localtime', '2011-01-03T04:21:22'),
                ('interface_list', [ 'FortyGigE1/0/1',
                                     'FortyGigE1/0/2',
                                     'FortyGigE1/0/3',
                                     'FortyGigE1/0/4',
                                     'FortyGigE1/0/5',
                                     'FortyGigE1/0/6',
                                     'FortyGigE1/0/7',
                                     'FortyGigE1/0/8',
                                     'FortyGigE1/0/9',
                                     'FortyGigE1/0/10',
                                     'FortyGigE1/0/11',
                                     'FortyGigE1/0/12',
                                     'FortyGigE1/0/13',
                                     'FortyGigE1/0/14',
                                     'FortyGigE1/0/15',
                                     'FortyGigE1/0/16',
                                     'FortyGigE1/0/17',
                                     'FortyGigE1/0/18',
                                     'FortyGigE1/0/19',
                                     'FortyGigE1/0/20',
                                     'FortyGigE1/0/21',
                                     'FortyGigE1/0/22',
                                     'FortyGigE1/0/23',
                                     'FortyGigE1/0/24',
                                     'FortyGigE1/0/25',
                                     'FortyGigE1/0/26',
                                     'FortyGigE1/0/27',
                                     'FortyGigE1/0/28',
                                     'FortyGigE1/0/29',
                                     'FortyGigE1/0/30',
                                     'FortyGigE1/0/31',
                                     'FortyGigE1/0/32',
                                     'M-GigabitEthernet0/0/0',
                                     'NULL0',
                                     'InLoopBack0',
                                     'Register-Tunnel0',
                                     'LoopBack29',
                                     'Bridge-Aggregation100'])
            ])

        self.assertEqual(dict(self.facts.facts), dict(expected))


if __name__ == '__main__':
    unittest.main()


HPDATA = 'http://www.hp.com/netconf/data:1.0'
HPDATA_C = '{' + HPDATA + '}'

NETCONFBASE = "urn:ietf:params:xml:ns:netconf:base:1.0"
NETCONFBASE_C = '{' + NETCONFBASE + '}'

HPCONFIG = "http://www.hp.com/netconf/config:1.0"
HPCONFIG_C = '{' + HPCONFIG + '}'

HPACTION = "http://www.hp.com/netconf/action:1.0"
HPACTION_C = '{' + HPACTION + '}'



import unittest
import mock

from pyhpecw7.features.l2vpn import L2VPN
from .base_feature_test import BaseFeatureCase

class L2VPNTestCase(BaseFeatureCase):

    @mock.patch('pyhpecw7.comware.HPCOM7')
    def setUp(self, mock_device):
        self.device = mock_device
        self.l2vpn = L2VPN(self.device)

    def test_get_config(self):
        expected_get, get_reply = self.xml_get_and_reply('l2vpn')
        self.device.get.return_value = get_reply

        expected = 'enabled'

        result = self.l2vpn.get_config()

        self.assertEqual(result, expected)
        self.assert_get_request(expected_get)

    def test_build_config(self):
        expected = self.read_config_xml('l2vpn_enabled')
        result = self.l2vpn._build_config('enabled')
        self.elements_are_equal(result, expected)

        expected = self.read_config_xml('l2vpn_disabled')
        result = self.l2vpn._build_config('disabled')
        self.elements_are_equal(result, expected)

    @mock.patch.object(L2VPN, '_build_config')
    def test_enable(self, mock_build_config):
        mock_build_config.return_value = self.read_config_xml('l2vpn_enabled')

        self.l2vpn.enable()
        mock_build_config.assert_called_with(state='enabled')
        self.assert_config_request(mock_build_config.return_value)

        self.l2vpn.enable(stage=True)
        mock_build_config.assert_called_with(state='enabled')
        self.assert_stage_request(mock_build_config.return_value, 'edit_config')

    @mock.patch.object(L2VPN, '_build_config')
    def test_disable(self, mock_build_config):
        mock_build_config.return_value = self.read_config_xml('l2vpn_disabled')

        self.l2vpn.disable()
        mock_build_config.assert_called_with(state='disabled')
        self.assert_config_request(mock_build_config.return_value)

        self.l2vpn.disable(stage=True)
        mock_build_config.assert_called_with(state='disabled')
        self.assert_stage_request(mock_build_config.return_value, 'edit_config')


if __name__ == "__main__":
    unittest.main()


import unittest
import os
from operator import attrgetter

from ncclient.operations.retrieve import GetReply, RPCReply
from lxml import etree

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))

class BaseFeatureCase(unittest.TestCase):

    def read_xml(self, operation, filename):
        if not filename.endswith('.xml'):
            filename = filename + '.xml'

        path = os.path.join(CURRENT_DIR, 'fixtures', operation, filename)
        with open(path) as f:
            content = f.read()

        return content

    def read_txt(self, operation, filename):
        if not filename.endswith('.txt'):
            filename = filename + '.txt'

        path = os.path.join(CURRENT_DIR, 'fixtures', operation, filename)
        with open(path) as f:
            content = f.read()

        return content

    def read_get_xml(self, filename):
        return etree.fromstring(self.read_xml('get', filename))

    def read_get_reply_xml(self, filename):
        return GetReply(self.read_xml('get_reply', filename))

    def read_config_xml(self, filename):
        return etree.fromstring(self.read_xml('edit_config', filename))

    def read_action_xml(self, filename):
        return etree.fromstring(self.read_xml('action', filename))

    def read_action_reply_xml(self, filename):
        return RPCReply(self.read_xml('action_reply', filename))

    def xml_get_and_reply(self, filename):
        return self.read_get_xml(filename), self.read_get_reply_xml(filename)

    def xml_action_and_reply(self, filename):
        return self.read_action_xml(filename), self.read_action_reply_xml(filename)

    def args_in_mock_call(self, func):
        last = len(func.mock_calls) - 1
        name, args, kwargs = func.mock_calls[last]
        while not args and last > 0:
            last = last - 1
            name, args, kwargs = func.mock_calls[last]

        return args

    def args_in_all_mock_calls(self, func):
        return list(x[1] for x in func.mock_calls if x[1])

    def read_cli_display(self, filename):
        return self.read_txt('cli_display', filename)

    def assert_get_request(self, expected):
        get_arg = self.args_in_mock_call(self.device.get)[0]

        self.assertEqual(get_arg[0], 'subtree')
        self.assert_elements_equal(get_arg[1], expected)

    def assert_action_request(self, expected):
        action_arg = self.args_in_mock_call(self.device.action)[0]
        self.assert_elements_equal(action_arg, expected)

    def assert_config_request(self, expected):
        config_arg = self.args_in_mock_call(self.device.edit_config)[0]
        self.assert_elements_equal(config_arg, expected)

    def assert_stage_request(self, expected, op_type):
        stage_args = self.args_in_mock_call(self.device.stage_config)

        self.assertEqual(stage_args[1], op_type)

        if isinstance(stage_args[0], str):
            self.assertEqual(stage_args[0], expected)
        else:
            self.assert_elements_equal(stage_args[0], expected)


    def assert_stage_requests_multiple(self, expected_list):
        actual_list = self.args_in_all_mock_calls(self.device.stage_config)
        for i in range(len(actual_list)):
            actual_ele, actual_op_type = tuple(actual_list[i])
            expected_ele, expected_op_type = tuple(expected_list[i])

            self.assertEqual(actual_op_type, expected_op_type)

            if isinstance(actual_ele, str):
                self.assertEqual(actual_ele, expected_ele)
            else:
                self.assert_elements_equal(actual_ele, expected_ele)

    def elements_are_equal(self, e1, e2):
        if e1.tag != e2.tag: return False
        if e1.text != e2.text:
            if e1.text == '' and e2.text is None:
                pass
            elif e1.text is None and e2.text == '':
                pass
            else:
                return False
        if e1.tail != e2.tail: return False
        if e1.attrib != e2.attrib: return False
        if len(e1) != len(e2): return False


        return all(self.elements_are_equal(c1, c2) for c1, c2 in zip(sorted(e1, key=attrgetter('tag')), sorted(e2, key=attrgetter('tag'))))

    def assert_elements_equal(self, e1, e2):
        are_equal = self.elements_are_equal(e1, e2)
        self.assertTrue(are_equal)


